# -*- coding: utf-8 -*-
# Copyright JS Foundation and other contributors, https://js.foundation/
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import, unicode_literals

import sys

import unicodedata
from collections import defaultdict

from .compat import uchr, xrange

# http://stackoverflow.com/questions/14245893/efficiently-list-all-characters-in-a-given-unicode-category
U_CATEGORIES = defaultdict(list)
for c in map(uchr, xrange(sys.maxunicode + 1)):
    U_CATEGORIES[unicodedata.category(c)].append(c)
UNICODE_LETTER = set(
    U_CATEGORIES['Lu'] + U_CATEGORIES['Ll'] +
    U_CATEGORIES['Lt'] + U_CATEGORIES['Lm'] +
    U_CATEGORIES['Lo'] + U_CATEGORIES['Nl']
)
UNICODE_OTHER_ID_START = set((
    # Other_ID_Start
    '\u1885', '\u1886', '\u2118', '\u212E', '\u309B', '\u309C',
    # New in Unicode 8.0
    '\u08B3', '\u0AF9', '\u13F8', '\u9FCD', '\uAB60', '\U00010CC0', '\U000108E0', '\U0002B820',
    # New in Unicode 9.0
    '\u1C80', '\U000104DB', '\U0001E922',
    '\U0001EE00', '\U0001EE06', '\U0001EE0A',
))
UNICODE_OTHER_ID_CONTINUE = set((
    # Other_ID_Continue
    '\xB7', '\u0387', '\u1369', '\u136A', '\u136B', '\u136C',
    '\u136D', '\u136E', '\u136F', '\u1370', '\u1371', '\u19DA',
    # New in Unicode 8.0
    '\u08E3', '\uA69E', '\U00011730',
    # New in Unicode 9.0
    '\u08D4', '\u1DFB', '\uA8C5', '\U00011450',
    '\U0001EE03', '\U0001EE0B',
))
UNICODE_COMBINING_MARK = set(U_CATEGORIES['Mn'] + U_CATEGORIES['Mc'])
UNICODE_DIGIT = set(U_CATEGORIES['Nd'])
UNICODE_CONNECTOR_PUNCTUATION = set(U_CATEGORIES['Pc'])
IDENTIFIER_START = UNICODE_LETTER.union(UNICODE_OTHER_ID_START).union(set(('$', '_', '\\')))
IDENTIFIER_PART = IDENTIFIER_START.union(UNICODE_COMBINING_MARK).union(UNICODE_DIGIT).union(UNICODE_CONNECTOR_PUNCTUATION).union(set(('\u200D', '\u200C'))).union(UNICODE_OTHER_ID_CONTINUE)

WHITE_SPACE = set((
    '\x09', '\x0B', '\x0C', '\x20', '\xA0',
    '\u1680', '\u180E', '\u2000', '\u2001', '\u2002',
    '\u2003', '\u2004', '\u2005', '\u2006', '\u2007',
    '\u2008', '\u2009', '\u200A', '\u202F', '\u205F',
    '\u3000', '\uFEFF',
))
LINE_TERMINATOR = set(('\x0A', '\x0D', '\u2028', '\u2029'))

DECIMAL_CONV = dict((c, n) for n, c in enumerate('0123456789'))
OCTAL_CONV = dict((c, n) for n, c in enumerate('01234567'))
HEX_CONV = dict((c, n) for n, c in enumerate('0123456789abcdef'))
for n, c in enumerate('ABCDEF', 10):
    HEX_CONV[c] = n
DECIMAL_DIGIT = set(DECIMAL_CONV.keys())
OCTAL_DIGIT = set(OCTAL_CONV.keys())
HEX_DIGIT = set(HEX_CONV.keys())


class Character:
    @staticmethod
    def fromCodePoint(code):
        return uchr(code)

    # https://tc39.github.io/ecma262/#sec-white-space

    @staticmethod
    def isWhiteSpace(ch):
        return ch in WHITE_SPACE

    # https://tc39.github.io/ecma262/#sec-line-terminators

    @staticmethod
    def isLineTerminator(ch):
        return ch in LINE_TERMINATOR

    # https://tc39.github.io/ecma262/#sec-names-and-keywords

    @staticmethod
    def isIdentifierStart(ch):
        return ch in IDENTIFIER_START

    @staticmethod
    def isIdentifierPart(ch):
        return ch in IDENTIFIER_PART

    # https://tc39.github.io/ecma262/#sec-literals-numeric-literals

    @staticmethod
    def isDecimalDigit(ch):
        return ch in DECIMAL_DIGIT

    @staticmethod
    def isHexDigit(ch):
        return ch in HEX_DIGIT

    @staticmethod
    def isOctalDigit(ch):
        return ch in OCTAL_DIGIT



import sys

from werkzeug.utils import cached_property
from werkzeug._compat import iteritems
from werkzeug._compat import to_unicode, to_bytes

import jam
from .common import consts, json_defaul_handler

class AbortException(Exception):
    pass

class AbstractItem(object):
    def __init__(self, task, owner, name='', caption='', visible = True, item_type_id=0, js_filename=''):
        self.task = task
        self.owner = owner
        self.item_name = name
        self.items = []
        self.ID = None
        self._events = []
        self.master = None
        self.js_filename = js_filename
        if owner:
            if not owner.find(name):
                owner.items.append(self)
                if not hasattr(owner, self.item_name):
                    setattr(owner, self.item_name, self)
        self.item_caption = caption
        self.visible = visible
        self.item_type_id = item_type_id
        if self != task:
            self.log = task.log
        self._loader = TracebackLoader(self)

    def task_locked(self):
        try:
            if self.task.ID:
                if self.master:
                    owner = self.master.owner
                else:
                    owner = self.owner
                if owner:
                    return self.task.app.task_locked()
        except:
            pass

    def __setattr__(self, name, value):
        if self.task_locked():
            raise Exception(self.task.language('server_tree_immutable') + \
                ' Item: "%s", Attribute: "%s"' % (self.item_name, name))
        super(AbstractItem, self).__setattr__(name, value)

    @property
    def session(self):
        try:
            return jam.context.session
        except:
            pass

    @property
    def environ(self):
        return jam.context.environ

    @property
    def user_info(self):
        if self.session:
            return self.session['user_info']

    def find(self, name):
        for item in self.items:
            if item.item_name == name:
                return item

    def item_by_ID(self, id_value):
        if self.ID == id_value:
            return self
        for item in self.items:
            result = item.item_by_ID(id_value)
            if result:
                return result

    def all(self, func):
        func(self);
        for item in self.items:
            item.all(func)

    def write_info(self, info, server):
        info['id'] = self.ID
        info['name'] = self.item_name
        info['caption'] = self.item_caption
        info['visible'] = self.visible
        info['type'] = self.item_type_id
        info['js_filename'] = self.js_filename

    def read_info(self, info):
        self.ID = info['id']
        self.item_name = info['name']
        self.item_caption = info['caption']
        self.visible = info['visible']
        self.item_type_id = info['type']
        self.js_filename = info['js_filename']

    def get_info(self, server=False):
        result = {}
        result['items'] = []
        self.write_info(result, server)
        for item in self.items:
            result['items'].append((item.item_type_id, item.get_info(server)))
        return result

    def get_child_class(self):
        pass

    def set_info(self, info):
        self.read_info(info)
        for item_type_id, item_info in info['items']:
            child = self.get_child_class()(self.task, self, item_info['name'])
            child.item_type_id = item_type_id
            child.set_info(item_info)

    def bind_item(self):
        pass

    def bind_items(self):
        self.bind_item()
        for item in self.items:
            item.bind_items()
        self.item_type = consts.ITEM_TYPES[self.item_type_id - 1]

    def get_module_name(self):
        result = self.owner.get_module_name() + '.' + self.item_name
        return str(result)

    def store_handlers(self):
        result = {}
        for key, value in iteritems(self.__dict__):
            if key[0:3] == 'on_':
                result[key] = self.__dict__[key]
        return result

    def clear_handlers(self):
        for key, value in iteritems(self.__dict__):
            if key[0:3] == 'on_':
                self.__dict__[key] = None

    def load_handlers(self, handlers):
        for key, value in iteritems(handlers):
            self.__dict__[key] = handlers[key]

    def get_master_field(self, fields, master_field):
        for field in fields:
            if field.ID == master_field:
                return field

    def abort(self, message=''):
        raise AbortException(message)

    def register(self, func):
        setattr(self, func.__name__, func)

    def load_code(self):
        return self.item_name

    def check_operation(self, operation):
        try:
            app = self.task.app
            if not consts.SAFE_MODE:
                return True
            elif self.task == app.admin:
                if consts.SAFE_MODE:
                    session = self.session
                    if session and session['user_info']['admin']:
                        return True
                else:
                    return True
            else:
                session = self.session
                if session:
                    role_id = session['user_info']['role_id']
                    privileges = self.task.app.get_privileges(role_id)
                    priv_dic = privileges.get(self.ID)
                    if priv_dic:
                        return priv_dic[operation]
        except:
            return False

    def can_view(self):
        return self.check_operation('can_view')


class AbstrGroup(AbstractItem):
    pass


class AbstrTask(AbstractItem):
    def __init__(self, owner, name, caption, visible = True, item_type_id=0, js_filename=''):
        AbstractItem.__init__(self, self, owner, name, caption, visible, item_type_id, js_filename)
        self.task = self
        self.item_type_id = consts.TASK_TYPE
        self.history_item = None
        self.lock_item = None
        self.log = None

    def task_locked(self):
        try:
            if self.ID:
                return self.app.task_locked()
        except:
            pass

    def write_info(self, info, server):
        super(AbstrTask, self).write_info(info, server)
        info['lookup_lists'] = self.lookup_lists
        if self.history_item:
            info['history_item'] = self.history_item.ID
        if self.lock_item:
            info['lock_item'] = self.lock_item.ID

    def set_info(self, info):
        super(AbstrTask, self).set_info(info)
        self.bind_items()

    def get_child_class(self):
        pass

    def item_by_name(self, item_name):
        for group in self.items:
            if group.item_name == item_name:
                return group
            else:
                for item in group.items:
                    if item.item_name == item_name:
                        return item

    def compile_item(self, item):
        pass

    def compile_all(self):
        for module in self.modules:
            del sys.modules[module]
        self.modules = []
        self.compile_item(self)
        for group in self.items:
            self.compile_item(group)
        for group in self.items:
            for item in group.items:
                self.compile_item(item)
        for group in self.items:
            for item in group.items:
                if group.item_type_id != consts.REPORTS_TYPE:
                    for detail in item.details:
                        self.compile_item(detail)

    def language(self, key):
        return consts.language(key)

class AbstrItem(AbstractItem):
    def __init__(self, task, owner, name, caption, visible = True, item_type_id=0, js_filename=''):
        AbstractItem.__init__(self, task, owner, name, caption, visible, item_type_id, js_filename)
        if not isinstance(self, AbstrDetail):
            if self.owner and not hasattr(self.task, self.item_name):
                setattr(self.task, self.item_name, self)

    def write_info(self, info, server):
        super(AbstrItem, self).write_info(info, server)
        info['fields'] = self.field_defs
        info['filters'] = self.filter_defs
        info['reports'] = self.get_reports_info()
        info['default_order'] = self._order_by
        info['primary_key'] = self._primary_key
        info['deleted_flag'] = self._deleted_flag
        info['virtual_table'] = self.virtual_table
        info['master_id'] = self._master_id
        info['master_rec_id'] = self._master_rec_id
        info['keep_history'] = self.keep_history
        info['edit_lock'] = self.edit_lock
        info['view_params'] = self._view_list
        info['edit_params'] = self._edit_list
        info['virtual_table'] = self.virtual_table
        if server:
            info['table_name'] = self.table_name

    def read_info(self, info):
        super(AbstrItem, self).read_info(info)
        self.create_fields(info['fields'])
        self.create_filters(info['filters'])
        self.reports = info['reports']
        self._order_by = info['default_order']
        self._primary_key = info['primary_key']
        self._deleted_flag = info['deleted_flag']
        self._virtual_table = info['virtual_table']
        self._master_id = info['master_id']
        self._master_rec_id = info['master_rec_id']
        self._keep_history = info['keep_history']
        self.edit_lock = info['edit_lock']
        self._view_list = info['view_params']
        self._edit_list = info['edit_params']
        self.table_name = info['table_name']

    def bind_item(self):
        self.prepare_fields()
        self.prepare_filters()

    def can_create(self):
        return self.check_operation('can_create')

    def can_edit(self):
        return self.check_operation('can_edit')

    def can_delete(self):
        return self.check_operation('can_delete')


class AbstrDetail(AbstrItem):

    def write_info(self, info, server):
        super(AbstrDetail, self).write_info(info, server)
        info['prototype_ID'] = self.prototype.ID

    def read_info(self, info):
        super(AbstrDetail, self).read_info(info)
        self.owner.details.append(self)
        if not hasattr(self.owner.details, self.item_name):
            setattr(self.owner.details, self.item_name, self)
        self._prototype_id = info['prototype_ID']

    def bind_item(self):
        super(AbstrDetail, self).bind_item();
        if hasattr(self, '_prototype_id'):
            self.prototype = self.task.item_by_ID(self._prototype_id)
            self.init_fields()

class AbstrReport(AbstractItem):
    def __init__(self, task, owner, name, caption, visible = True, item_type_id=0, js_filename=''):
        AbstractItem.__init__(self, task, owner, name, caption, visible, item_type_id, js_filename)
        if not hasattr(self.task, self.item_name):
            setattr(self.task, self.item_name, self)

    def write_info(self, info, server):
        super(AbstrReport, self).write_info(info, server)
        info['fields'] = self.param_defs

    def read_info(self, info):
        super(AbstrReport, self).read_info(info)
        self.create_params(info['fields'])

    def param_by_name(self, name):
        for param in self.params:
            if param.param_name == name:
                return param

    def bind_item(self):
        self.prepare_params()

class TracebackLoader(object):
    def __init__(self, item):
        self.item = item

    def get_source(self, source):
        admin = self.item.task.app.admin
        sys_items = admin.sys_items.copy()
        sys_items.set_where(id=self.item.ID)
        sys_items.open(fields=['f_server_module'])
        return sys_items.f_server_module.value



# -*- coding: utf-8 -*-
# Copyright JS Foundation and other contributors, https://js.foundation/
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import unicode_literals


# Error messages should be identical to V8.
class Messages:
    ObjectPatternAsRestParameter = "Unexpected token {"
    BadImportCallArity = "Unexpected token"
    BadGetterArity = "Getter must not have any formal parameters"
    BadSetterArity = "Setter must have exactly one formal parameter"
    BadSetterRestParameter = "Setter function argument must not be a rest parameter"
    ConstructorIsAsync = "Class constructor may not be an async method"
    ConstructorSpecialMethod = "Class constructor may not be an accessor"
    DeclarationMissingInitializer = "Missing initializer in %0 declaration"
    DefaultRestParameter = "Unexpected token ="
    DefaultRestProperty = "Unexpected token ="
    DuplicateBinding = "Duplicate binding %0"
    DuplicateConstructor = "A class may only have one constructor"
    DuplicateProtoProperty = "Duplicate __proto__ fields are not allowed in object literals"
    ForInOfLoopInitializer = "%0 loop variable declaration may not have an initializer"
    GeneratorInLegacyContext = "Generator declarations are not allowed in legacy contexts"
    IllegalBreak = "Illegal break statement"
    IllegalContinue = "Illegal continue statement"
    IllegalExportDeclaration = "Unexpected token"
    IllegalImportDeclaration = "Unexpected token"
    IllegalLanguageModeDirective = "Illegal 'use strict' directive in function with non-simple parameter list"
    IllegalReturn = "Illegal return statement"
    InvalidEscapedReservedWord = "Keyword must not contain escaped characters"
    InvalidHexEscapeSequence = "Invalid hexadecimal escape sequence"
    InvalidLHSInAssignment = "Invalid left-hand side in assignment"
    InvalidLHSInForIn = "Invalid left-hand side in for-in"
    InvalidLHSInForLoop = "Invalid left-hand side in for-loop"
    InvalidModuleSpecifier = "Unexpected token"
    InvalidRegExp = "Invalid regular expression"
    LetInLexicalBinding = "let is disallowed as a lexically bound name"
    MissingFromClause = "Unexpected token"
    MultipleDefaultsInSwitch = "More than one default clause in switch statement"
    NewlineAfterThrow = "Illegal newline after throw"
    NoAsAfterImportNamespace = "Unexpected token"
    NoCatchOrFinally = "Missing catch or finally after try"
    ParameterAfterRestParameter = "Rest parameter must be last formal parameter"
    PropertyAfterRestProperty = "Unexpected token"
    Redeclaration = "%0 '%1' has already been declared"
    StaticPrototype = "Classes may not have static property named prototype"
    StrictCatchVariable = "Catch variable may not be eval or arguments in strict mode"
    StrictDelete = "Delete of an unqualified identifier in strict mode."
    StrictFunction = "In strict mode code, functions can only be declared at top level or inside a block"
    StrictFunctionName = "Function name may not be eval or arguments in strict mode"
    StrictLHSAssignment = "Assignment to eval or arguments is not allowed in strict mode"
    StrictLHSPostfix = "Postfix increment/decrement may not have eval or arguments operand in strict mode"
    StrictLHSPrefix = "Prefix increment/decrement may not have eval or arguments operand in strict mode"
    StrictModeWith = "Strict mode code may not include a with statement"
    StrictOctalLiteral = "Octal literals are not allowed in strict mode."
    StrictParamDupe = "Strict mode function may not have duplicate parameter names"
    StrictParamName = "Parameter name eval or arguments is not allowed in strict mode"
    StrictReservedWord = "Use of future reserved word in strict mode"
    StrictVarName = "Variable name may not be eval or arguments in strict mode"
    TemplateOctalLiteral = "Octal literals are not allowed in template strings."
    UnexpectedEOS = "Unexpected end of input"
    UnexpectedIdentifier = "Unexpected identifier"
    UnexpectedNumber = "Unexpected number"
    UnexpectedReserved = "Unexpected reserved word"
    UnexpectedString = "Unexpected string"
    UnexpectedTemplate = "Unexpected quasi %0"
    UnexpectedToken = "Unexpected token %0"
    UnexpectedTokenIllegal = "Unexpected token ILLEGAL"
    UnknownLabel = "Undefined label '%0'"
    UnterminatedRegExp = "Invalid regular expression: missing /"



"""
werkzeug
~~~~~~~~

Werkzeug is the Swiss Army knife of Python web development.

It provides useful classes and functions for any WSGI application to
make the life of a Python web developer much easier. All of the provided
classes are independent from each other so you can mix it with any other
library.

:copyright: 2007 Pallets
:license: BSD-3-Clause
"""
from .serving import run_simple
from .test import Client
from .wrappers import Request
from .wrappers import Response

__version__ = "1.0.0"



# -*- coding: utf-8 -*-
# Copyright JS Foundation and other contributors, https://js.foundation/
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import unicode_literals

import json
import types
from collections import deque

from .objects import Object
from .compat import PY3, unicode


class VisitRecursionError(Exception):
    pass


class Visited(object):
    def __init__(self, result):
        if isinstance(result, Visited):
            result = result.result
        self.result = result


class Visitor(object):
    """
    An Object visitor base class that walks the abstract syntax tree and calls a
    visitor function for every Object found.  This function may return a value
    which is forwarded by the `visit` method.

    This class is meant to be subclassed, with the subclass adding visitor
    methods.

    Per default the visitor functions for the nodes are ``'visit_'`` +
    class name of the Object.  So a `Module` Object visit function would
    be `visit_Module`.  This behavior can be changed by overriding
    the `visit` method.  If no visitor function exists for a Object
    (return value `None`) the `generic_visit` visitor is used instead.
    """

    def __call__(self, obj, metadata):
        return self.transform(obj, metadata)

    def transform(self, obj, metadata):
        """Transform an Object."""
        if isinstance(obj, Object):
            method = 'transform_' + obj.__class__.__name__
            transformer = getattr(self, method, self.transform_Object)
            new_obj = transformer(obj, metadata)
            if new_obj is not None and obj is not new_obj:
                obj = new_obj
        return obj

    def transform_Object(self, obj, metadata):
        """Called if no explicit transform function exists for an Object."""
        return obj

    def generic_visit(self, obj):
        return self.visit(self.visit_Object(obj))

    def visit(self, obj):
        """Visit a Object."""
        if not hasattr(self, 'visitors'):
            self._visit_context = {}
            self._visit_count = 0
        try:
            self._visit_count += 1
            stack = deque()
            stack.append((obj, None))
            last_result = None
            while stack:
                try:
                    last, visited = stack[-1]
                    if isinstance(last, types.GeneratorType):
                        stack.append((last.send(last_result), None))
                        last_result = None
                    elif isinstance(last, Visited):
                        stack.pop()
                        last_result = last.result
                    elif isinstance(last, Object):
                        if last in self._visit_context:
                            if self._visit_context[last] == self.visit_Object:
                                visitor = self.visit_RecursionError
                            else:
                                visitor = self.visit_Object
                        else:
                            method = 'visit_' + last.__class__.__name__
                            visitor = getattr(self, method, self.visit_Object)
                        self._visit_context[last] = visitor
                        stack.pop()
                        stack.append((visitor(last), last))
                    else:
                        method = 'visit_' + last.__class__.__name__
                        visitor = getattr(self, method, self.visit_Generic)
                        stack.pop()
                        stack.append((visitor(last), None))
                except StopIteration:
                    stack.pop()
                    if visited and visited in self._visit_context:
                        del self._visit_context[visited]
            return last_result
        finally:
            self._visit_count -= 1
            if self._visit_count <= 0:
                self._visit_context = {}

    def visit_RecursionError(self, obj):
        raise VisitRecursionError

    def visit_Object(self, obj):
        """Called if no explicit visitor function exists for an Object."""
        yield obj.__dict__
        yield Visited(obj)

    def visit_Generic(self, obj):
        """Called if no explicit visitor function exists for an object."""
        yield Visited(obj)

    def visit_list(self, obj):
        for item in obj:
            yield item
        yield Visited(obj)

    visit_Array = visit_list

    def visit_dict(self, obj):
        for field, value in list(obj.items()):
            if not field.startswith('_'):
                yield value
        yield Visited(obj)


class NodeVisitor(Visitor):
    pass


class ReprVisitor(Visitor):
    def visit(self, obj, indent=4, nl="\n", sp="", skip=()):
        self.level = 0
        if isinstance(indent, int):
            indent = " " * indent
        self.indent = indent
        self.nl = nl
        self.sp = sp
        self.skip = skip
        return super(ReprVisitor, self).visit(obj)

    def visit_RecursionError(self, obj):
        yield Visited("...")

    def visit_Object(self, obj):
        value_repr = yield obj.__dict__
        yield Visited(value_repr)

    def visit_Generic(self, obj):
        yield Visited(repr(obj))

    def visit_list(self, obj):
        indent1 = self.indent * self.level
        indent2 = indent1 + self.indent
        self.level += 1
        try:
            items = []
            for item in obj:
                v = yield item
                items.append(v)
            if items:
                value_repr = "[%s%s%s%s%s%s%s]" % (
                    self.sp,
                    self.nl,
                    indent2,
                    (",%s%s%s" % (self.nl, self.sp, indent2)).join(items),
                    self.nl,
                    indent1,
                    self.sp,
                )
            else:
                value_repr = "[]"
        finally:
            self.level -= 1

        yield Visited(value_repr)

    visit_Array = visit_list

    def visit_dict(self, obj):
        indent1 = self.indent * self.level
        indent2 = indent1 + self.indent
        self.level += 1
        try:
            items = []
            for k, item in obj.items():
                if item is not None and not k.startswith('_') and k not in self.skip:
                    v = yield item
                    items.append("%s: %s" % (k, v))
            if items:
                value_repr = "{%s%s%s%s%s%s%s}" % (
                    self.sp,
                    self.nl,
                    indent2,
                    (",%s%s%s" % (self.nl, self.sp, indent2)).join(items),
                    self.nl,
                    indent1,
                    self.sp,
                )
            else:
                value_repr = "{}"
        finally:
            self.level -= 1

        yield Visited(value_repr)

    if PY3:
        def visit_str(self, obj):
            value_repr = json.dumps(obj)
            yield Visited(value_repr)
    else:
        def visit_unicode(self, obj):
            value_repr = json.dumps(obj)
            yield Visited(value_repr)

    def visit_SourceLocation(self, obj):
        old_indent, self.indent = self.indent, ""
        old_nl, self.nl = self.nl, ""
        old_sp, self.sp = self.sp, ""
        try:
            yield obj
        finally:
            self.indent = old_indent
            self.nl = old_nl
            self.sp = old_sp


class ToDictVisitor(Visitor):
    map = {
        'isAsync': 'async',
        'allowAwait': 'await',
    }

    def visit_RecursionError(self, obj):
        yield Visited({
            'error': "Infinite recursion detected...",
        })

    def visit_Object(self, obj):
        obj = yield obj.__dict__
        yield Visited(obj)

    def visit_list(self, obj):
        items = []
        for item in obj:
            v = yield item
            items.append(v)
        yield Visited(items)

    visit_Array = visit_list

    def visit_dict(self, obj):
        items = []
        for k, item in obj.items():
            if item is not None and not k.startswith('_'):
                v = yield item
                k = unicode(k)
                items.append((self.map.get(k, k), v))
        yield Visited(dict(items))

    def visit_SRE_Pattern(self, obj):
        yield Visited({})



# -*- coding: utf-8 -*-
# Copyright JS Foundation and other contributors, https://js.foundation/
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import, unicode_literals

from .compat import uchr
from .character import Character
from . import jsx_nodes as JSXNode
from .jsx_syntax import JSXSyntax
from . import nodes as Node
from .parser import Marker, Parser
from .token import Token, TokenName
from .xhtml_entities import XHTMLEntities


class MetaJSXElement(object):
    def __init__(self, node=None, opening=None, closing=None, children=None):
        self.node = node
        self.opening = opening
        self.closing = closing
        self.children = children


class JSXToken(object):
    Identifier = 100
    Text = 101


class RawJSXToken(object):
    def __init__(self, type=None, value=None, lineNumber=None, lineStart=None, start=None, end=None):
        self.type = type
        self.value = value
        self.lineNumber = lineNumber
        self.lineStart = lineStart
        self.start = start
        self.end = end


TokenName[JSXToken.Identifier] = "JSXIdentifier"
TokenName[JSXToken.Text] = "JSXText"


# Fully qualified element name, e.g. <svg:path> returns "svg:path"
def getQualifiedElementName(elementName):
    typ = elementName.type
    if typ is JSXSyntax.JSXIdentifier:
        id = elementName
        qualifiedName = id.name
    elif typ is JSXSyntax.JSXNamespacedName:
        ns = elementName
        qualifiedName = getQualifiedElementName(ns.namespace) + ':' + getQualifiedElementName(ns.name)
    elif typ is JSXSyntax.JSXMemberExpression:
        expr = elementName
        qualifiedName = getQualifiedElementName(expr.object) + '.' + getQualifiedElementName(expr.property)

    return qualifiedName


class JSXParser(Parser):
    def __init__(self, code, options, delegate):
        super(JSXParser, self).__init__(code, options, delegate)

    def parsePrimaryExpression(self):
        return self.parseJSXRoot() if self.match('<') else super(JSXParser, self).parsePrimaryExpression()

    def startJSX(self):
        # Unwind the scanner before the lookahead token.
        self.scanner.index = self.startMarker.index
        self.scanner.lineNumber = self.startMarker.line
        self.scanner.lineStart = self.startMarker.index - self.startMarker.column

    def finishJSX(self):
        # Prime the next lookahead.
        self.nextToken()

    def reenterJSX(self):
        self.startJSX()
        self.expectJSX('}')

        # Pop the closing '}' added from the lookahead.
        if self.config.tokens:
            self.tokens.pop()

    def createJSXNode(self):
        self.collectComments()
        return Marker(
            index=self.scanner.index,
            line=self.scanner.lineNumber,
            column=self.scanner.index - self.scanner.lineStart
        )

    def createJSXChildNode(self):
        return Marker(
            index=self.scanner.index,
            line=self.scanner.lineNumber,
            column=self.scanner.index - self.scanner.lineStart
        )

    def scanXHTMLEntity(self, quote):
        result = '&'

        valid = True
        terminated = False
        numeric = False
        hex = False

        while not self.scanner.eof() and valid and not terminated:
            ch = self.scanner.source[self.scanner.index]
            if ch == quote:
                break

            terminated = (ch == ';')
            result += ch
            self.scanner.index += 1
            if not terminated:
                length = len(result)
                if length == 2:
                    # e.g. '&#123;'
                    numeric = (ch == '#')
                elif length == 3:
                    if numeric:
                        # e.g. '&#x41;'
                        hex = ch == 'x'
                        valid = hex or Character.isDecimalDigit(ch)
                        numeric = numeric and not hex
                else:
                    valid = valid and not (numeric and not Character.isDecimalDigit(ch))
                    valid = valid and not (hex and not Character.isHexDigit(ch))

        if valid and terminated and len(result) > 2:
            # e.g. '&#x41;' becomes just '#x41'
            st = result[1:-1]
            if numeric and len(st) > 1:
                result = uchr(int(st[1:], 10))
            elif hex and len(st) > 2:
                result = uchr(int(st[2:], 16))
            elif not numeric and not hex and st in XHTMLEntities:
                result = XHTMLEntities[st]

        return result

    # Scan the next JSX token. This replaces Scanner#lex when in JSX mode.

    def lexJSX(self):
        ch = self.scanner.source[self.scanner.index]

        # < > / : = { }
        if ch in ('<', '>', '/', ':', '=', '{', '}'):
            value = self.scanner.source[self.scanner.index]
            self.scanner.index += 1
            return RawJSXToken(
                type=Token.Punctuator,
                value=value,
                lineNumber=self.scanner.lineNumber,
                lineStart=self.scanner.lineStart,
                start=self.scanner.index - 1,
                end=self.scanner.index
            )

        # " '
        if ch in ('\'', '"'):
            start = self.scanner.index
            quote = self.scanner.source[self.scanner.index]
            self.scanner.index += 1
            str = ''
            while not self.scanner.eof():
                ch = self.scanner.source[self.scanner.index]
                self.scanner.index += 1
                if ch == quote:
                    break
                elif ch == '&':
                    str += self.scanXHTMLEntity(quote)
                else:
                    str += ch

            return RawJSXToken(
                type=Token.StringLiteral,
                value=str,
                lineNumber=self.scanner.lineNumber,
                lineStart=self.scanner.lineStart,
                start=start,
                end=self.scanner.index
            )

        # ... or .
        if ch == '.':
            start = self.scanner.index
            if self.scanner.source[start + 1:start + 3] == '..':
                value = '...'
                self.scanner.index += 3
            else:
                value = '.'
                self.scanner.index += 1
            return RawJSXToken(
                type=Token.Punctuator,
                value=value,
                lineNumber=self.scanner.lineNumber,
                lineStart=self.scanner.lineStart,
                start=start,
                end=self.scanner.index
            )

        # `
        if ch == '`':
            # Only placeholder, since it will be rescanned as a real assignment expression.
            return RawJSXToken(
                type=Token.Template,
                value='',
                lineNumber=self.scanner.lineNumber,
                lineStart=self.scanner.lineStart,
                start=self.scanner.index,
                end=self.scanner.index
            )

        # Identifer can not contain backslash (char code 92).
        if Character.isIdentifierStart(ch) and ch != '\\':
            start = self.scanner.index
            self.scanner.index += 1
            while not self.scanner.eof():
                ch = self.scanner.source[self.scanner.index]
                if Character.isIdentifierPart(ch) and ch != '\\':
                    self.scanner.index += 1
                elif ch == '-':
                    # Hyphen (char code 45) can be part of an identifier.
                    self.scanner.index += 1
                else:
                    break

            id = self.scanner.source[start:self.scanner.index]
            return RawJSXToken(
                type=JSXToken.Identifier,
                value=id,
                lineNumber=self.scanner.lineNumber,
                lineStart=self.scanner.lineStart,
                start=start,
                end=self.scanner.index
            )

        return self.scanner.lex()

    def nextJSXToken(self):
        self.collectComments()

        self.startMarker.index = self.scanner.index
        self.startMarker.line = self.scanner.lineNumber
        self.startMarker.column = self.scanner.index - self.scanner.lineStart
        token = self.lexJSX()
        self.lastMarker.index = self.scanner.index
        self.lastMarker.line = self.scanner.lineNumber
        self.lastMarker.column = self.scanner.index - self.scanner.lineStart

        if self.config.tokens:
            self.tokens.append(self.convertToken(token))

        return token

    def nextJSXText(self):
        self.startMarker.index = self.scanner.index
        self.startMarker.line = self.scanner.lineNumber
        self.startMarker.column = self.scanner.index - self.scanner.lineStart

        start = self.scanner.index

        text = ''
        while not self.scanner.eof():
            ch = self.scanner.source[self.scanner.index]
            if ch in ('{', '<'):
                break

            self.scanner.index += 1
            text += ch
            if Character.isLineTerminator(ch):
                self.scanner.lineNumber += 1
                if ch == '\r' and self.scanner.source[self.scanner.index] == '\n':
                    self.scanner.index += 1

                self.scanner.lineStart = self.scanner.index

        self.lastMarker.index = self.scanner.index
        self.lastMarker.line = self.scanner.lineNumber
        self.lastMarker.column = self.scanner.index - self.scanner.lineStart

        token = RawJSXToken(
            type=JSXToken.Text,
            value=text,
            lineNumber=self.scanner.lineNumber,
            lineStart=self.scanner.lineStart,
            start=start,
            end=self.scanner.index
        )

        if text and self.config.tokens:
            self.tokens.append(self.convertToken(token))

        return token

    def peekJSXToken(self):
        state = self.scanner.saveState()
        self.scanner.scanComments()
        next = self.lexJSX()
        self.scanner.restoreState(state)

        return next

    # Expect the next JSX token to match the specified punctuator.
    # If not, an exception will be thrown.

    def expectJSX(self, value):
        token = self.nextJSXToken()
        if token.type is not Token.Punctuator or token.value != value:
            self.throwUnexpectedToken(token)

    # Return True if the next JSX token matches the specified punctuator.

    def matchJSX(self, *value):
        next = self.peekJSXToken()
        return next.type is Token.Punctuator and next.value in value

    def parseJSXIdentifier(self):
        node = self.createJSXNode()
        token = self.nextJSXToken()
        if token.type is not JSXToken.Identifier:
            self.throwUnexpectedToken(token)

        return self.finalize(node, JSXNode.JSXIdentifier(token.value))

    def parseJSXElementName(self):
        node = self.createJSXNode()
        elementName = self.parseJSXIdentifier()

        if self.matchJSX(':'):
            namespace = elementName
            self.expectJSX(':')
            name = self.parseJSXIdentifier()
            elementName = self.finalize(node, JSXNode.JSXNamespacedName(namespace, name))
        elif self.matchJSX('.'):
            while self.matchJSX('.'):
                object = elementName
                self.expectJSX('.')
                property = self.parseJSXIdentifier()
                elementName = self.finalize(node, JSXNode.JSXMemberExpression(object, property))

        return elementName

    def parseJSXAttributeName(self):
        node = self.createJSXNode()

        identifier = self.parseJSXIdentifier()
        if self.matchJSX(':'):
            namespace = identifier
            self.expectJSX(':')
            name = self.parseJSXIdentifier()
            attributeName = self.finalize(node, JSXNode.JSXNamespacedName(namespace, name))
        else:
            attributeName = identifier

        return attributeName

    def parseJSXStringLiteralAttribute(self):
        node = self.createJSXNode()
        token = self.nextJSXToken()
        if token.type is not Token.StringLiteral:
            self.throwUnexpectedToken(token)

        raw = self.getTokenRaw(token)
        return self.finalize(node, Node.Literal(token.value, raw))

    def parseJSXExpressionAttribute(self):
        node = self.createJSXNode()

        self.expectJSX('{')
        self.finishJSX()

        if self.match('}'):
            self.tolerateError('JSX attributes must only be assigned a non-empty expression')

        expression = self.parseAssignmentExpression()
        self.reenterJSX()

        return self.finalize(node, JSXNode.JSXExpressionContainer(expression))

    def parseJSXAttributeValue(self):
        if self.matchJSX('{'):
            return self.parseJSXExpressionAttribute()
        if self.matchJSX('<'):
            return self.parseJSXElement()

        return self.parseJSXStringLiteralAttribute()

    def parseJSXNameValueAttribute(self):
        node = self.createJSXNode()
        name = self.parseJSXAttributeName()
        value = None
        if self.matchJSX('='):
            self.expectJSX('=')
            value = self.parseJSXAttributeValue()

        return self.finalize(node, JSXNode.JSXAttribute(name, value))

    def parseJSXSpreadAttribute(self):
        node = self.createJSXNode()
        self.expectJSX('{')
        self.expectJSX('...')

        self.finishJSX()
        argument = self.parseAssignmentExpression()
        self.reenterJSX()

        return self.finalize(node, JSXNode.JSXSpreadAttribute(argument))

    def parseJSXAttributes(self):
        attributes = []

        while not self.matchJSX('/', '>'):
            attribute = self.parseJSXSpreadAttribute() if self.matchJSX('{') else self.parseJSXNameValueAttribute()
            attributes.append(attribute)

        return attributes

    def parseJSXOpeningElement(self):
        node = self.createJSXNode()

        self.expectJSX('<')
        name = self.parseJSXElementName()
        attributes = self.parseJSXAttributes()
        selfClosing = self.matchJSX('/')
        if selfClosing:
            self.expectJSX('/')

        self.expectJSX('>')

        return self.finalize(node, JSXNode.JSXOpeningElement(name, selfClosing, attributes))

    def parseJSXBoundaryElement(self):
        node = self.createJSXNode()

        self.expectJSX('<')
        if self.matchJSX('/'):
            self.expectJSX('/')
            elementName = self.parseJSXElementName()
            self.expectJSX('>')
            return self.finalize(node, JSXNode.JSXClosingElement(elementName))

        name = self.parseJSXElementName()
        attributes = self.parseJSXAttributes()
        selfClosing = self.matchJSX('/')
        if selfClosing:
            self.expectJSX('/')

        self.expectJSX('>')

        return self.finalize(node, JSXNode.JSXOpeningElement(name, selfClosing, attributes))

    def parseJSXEmptyExpression(self):
        node = self.createJSXChildNode()
        self.collectComments()
        self.lastMarker.index = self.scanner.index
        self.lastMarker.line = self.scanner.lineNumber
        self.lastMarker.column = self.scanner.index - self.scanner.lineStart
        return self.finalize(node, JSXNode.JSXEmptyExpression())

    def parseJSXExpressionContainer(self):
        node = self.createJSXNode()
        self.expectJSX('{')

        if self.matchJSX('}'):
            expression = self.parseJSXEmptyExpression()
            self.expectJSX('}')
        else:
            self.finishJSX()
            expression = self.parseAssignmentExpression()
            self.reenterJSX()

        return self.finalize(node, JSXNode.JSXExpressionContainer(expression))

    def parseJSXChildren(self):
        children = []

        while not self.scanner.eof():
            node = self.createJSXChildNode()
            token = self.nextJSXText()
            if token.start < token.end:
                raw = self.getTokenRaw(token)
                child = self.finalize(node, JSXNode.JSXText(token.value, raw))
                children.append(child)

            if self.scanner.source[self.scanner.index] == '{':
                container = self.parseJSXExpressionContainer()
                children.append(container)
            else:
                break

        return children

    def parseComplexJSXElement(self, el):
        stack = []

        while not self.scanner.eof():
            el.children.extend(self.parseJSXChildren())
            node = self.createJSXChildNode()
            element = self.parseJSXBoundaryElement()
            if element.type is JSXSyntax.JSXOpeningElement:
                opening = element
                if opening.selfClosing:
                    child = self.finalize(node, JSXNode.JSXElement(opening, [], None))
                    el.children.append(child)
                else:
                    stack.append(el)
                    el = MetaJSXElement(
                        node=node,
                        opening=opening,
                        closing=None,
                        children=[],
                    )

            if element.type is JSXSyntax.JSXClosingElement:
                el.closing = element
                open = getQualifiedElementName(el.opening.name)
                close = getQualifiedElementName(el.closing.name)
                if open != close:
                    self.tolerateError('Expected corresponding JSX closing tag for %0', open)

                if stack:
                    child = self.finalize(el.node, JSXNode.JSXElement(el.opening, el.children, el.closing))
                    el = stack[-1]
                    el.children.append(child)
                    stack.pop()
                else:
                    break

        return el

    def parseJSXElement(self):
        node = self.createJSXNode()

        opening = self.parseJSXOpeningElement()
        children = []
        closing = None

        if not opening.selfClosing:
            el = self.parseComplexJSXElement(MetaJSXElement(
                node=node,
                opening=opening,
                closing=closing,
                children=children
            ))
            children = el.children
            closing = el.closing

        return self.finalize(node, JSXNode.JSXElement(opening, children, closing))

    def parseJSXRoot(self):
        # Pop the opening '<' added from the lookahead.
        if self.config.tokens:
            self.tokens.pop()

        self.startJSX()
        element = self.parseJSXElement()
        self.finishJSX()

        return element

    def isStartOfExpression(self):
        return super(JSXParser, self).isStartOfExpression() or self.match('<')



import os
import json
import sqlite3

from ..common import consts, error_message, file_read, file_write
from ..server_classes import AdminTask, Group
from .builder import on_created, init_task_attr
from jam.db.db_modules import SQLITE, get_db_module
import jam.langs as langs

def create_items(task):
    info = file_read(os.path.join(task.app.jam_dir, 'admin', 'builder_structure.info'))
    info = json.loads(info)
    task.set_info(info)

def read_secret_key(task):
    result = None
    con = task.connect()
    try:
        cursor = con.cursor()
        cursor.execute('SELECT F_SECRET_KEY FROM SYS_PARAMS')
        rec = cursor.fetchall()
        result = rec[0][0]
    except:
        pass
    finally:
        con.close()
    if result is None:
        result = ''
    return result

def check_version(task):
    pass

def init_admin(task):
    check_version(task)
    langs.update_langs(task)
    create_items(task)
    update_admin_fields(task)
    consts.read_settings()
    consts.MAINTENANCE = False
    consts.write_settings(['MAINTENANCE'])
    consts.read_language()
    on_created(task)

def create_admin(app):
    if os.path.exists(os.path.join(app.work_dir, '_admin.sqlite')):
        os.rename(os.path.join(app.work_dir, '_admin.sqlite'), \
            os.path.join(app.work_dir, 'admin.sqlite'))
    task = AdminTask(app, 'admin', 'Administrator', '', SQLITE,
        db_database=os.path.join(app.work_dir, 'admin.sqlite'))
    app.admin = task
    task.secret_key = read_secret_key(task)
    init_admin(task)
    return task

def update_admin_fields(task):

    def do_updates(con, field, item_name):
        if item_name == 'sys_privileges' and field.field_name.lower() == 'owner_item':
            cursor = con.cursor()
            cursor.execute("SELECT ID FROM SYS_ITEMS WHERE TABLE_ID > 0 AND DELETED = 0")
            details = cursor.fetchall()
            cursor.execute("SELECT ID FROM SYS_ROLES WHERE DELETED = 0")
            roles = cursor.fetchall()
            for d in details:
                for r in roles:
                    cursor.execute("""
                        INSERT INTO SYS_PRIVILEGES
                        (DELETED, OWNER_ID, OWNER_REC_ID, ITEM_ID, F_CAN_VIEW, F_CAN_CREATE, F_CAN_EDIT, F_CAN_DELETE)
                        values (?, ?, ?, ?, ?, ?, ?, ?)""",
                        (0, 2, r[0], d[0], True, True, True, True))
            con.commit()

    def get_item_fields(item, table_name):
        cursor.execute('PRAGMA table_info(%s)' % table_name)
        rows = cursor.fetchall()
        result = [str(row[1]).upper() for row in rows]
        return result

    def check_item_fields(item, table_name=None):
        if not table_name:
            table_name = item.table_name.upper()
        fields = get_item_fields(item, table_name)
        for field in item._fields:
            if not field.field_name.upper() in fields:
                sql = 'ALTER TABLE %s ADD COLUMN %s %s' % \
                    (table_name, field.field_name.upper(), \
                    task.db_module.FIELD_TYPES[field.data_type])
                cursor.execute(sql)
                con.commit()
                do_updates(con, field, item.item_name)

    def check_table_exists(item, table_name=None):
        if not table_name:
            table_name = item.table_name.upper()
        sql = 'SELECT name FROM sqlite_master WHERE type="table" AND UPPER(name)="%s"' % table_name
        cursor.execute(sql)
        rows = cursor.fetchall()
        if not rows:
            sql = 'CREATE TABLE %s (ID INTEGER PRIMARY KEY)' % table_name
            cursor.execute(sql)
        return True

    con = task.connect()
    try:
        cursor = con.cursor()
        for group in task.items:
            for item in group.items:
                if item.table_name and not item.master:
                    if check_table_exists(item):
                        check_item_fields(item)
    finally:
        con.close()

def get_privileges(task, role_id):
    result = {}
    privliges = task.sys_privileges.copy()
    privliges.set_where(owner_rec_id=role_id)
    privliges.open()
    for p in privliges:
        result[p.item_id.value] = \
            {
            'can_view': p.f_can_view.value,
            'can_create': p.f_can_create.value,
            'can_edit': p.f_can_edit.value,
            'can_delete': p.f_can_delete.value
            }
    return result

def get_roles(task):
    privileges = {}
    roles = []
    r = task.sys_roles.copy()
    r.open()
    for r in r:
        privileges[r.id.value] = get_privileges(task, r.id.value)
        roles.append([r.id.value, r.f_name.value])
    return roles, privileges

def login_user(task, log, password, admin, ip=None, session_uuid=None):
    user_id = None
    user_info = {}
    if consts.SAFE_MODE:
        users = task.sys_users.copy()
        users.set_where(f_password=password)
        users.open()
        for u in users:
            if u.f_login.value.strip() == log.strip() and u.f_password.value == password:
                if not admin or u.f_admin.value == admin:
                    user_id = u.id.value
                    user_info = {
                        'user_id': u.id.value,
                        'role_id': u.f_role.value,
                        'role_name': u.f_role.display_text,
                        'user_name': u.f_name.value,
                        'admin': u.f_admin.value
                    }
                    if ip or session_uuid:
                        task.execute("UPDATE SYS_USERS SET F_IP='%s', F_UUID='%s' WHERE ID=%s" % (ip, session_uuid, u.id.value))
                    break
    return user_info

def user_valid_ip(task, user_id, ip):
    res = task.select("SELECT F_IP FROM SYS_USERS WHERE ID=%s" % user_id)
    if res and res[0][0] == ip:
        return True
    return False

def user_valid_uuid(task, user_id, session_uuid):
    res = task.select("SELECT F_UUID FROM SYS_USERS WHERE ID=%s" % user_id)
    if res and res[0][0] == session_uuid:
        return True
    return False

def indexes_get_table_names(indexes):
    ids = []
    for i in indexes:
        ids.append(i.owner_rec_id.value)
    items = indexes.task.sys_items.copy(handlers=False)
    items.set_where(id__in=ids)
    items.open(fields=['id', 'f_table_name'])
    table_names = {}
    for i in items:
        table_names[i.id.value] = i.f_table_name.value
    return table_names

def drop_indexes_sql(task):
    db_module = task.task_db_module
    db_type = task.task_db_type
    indexes = task.sys_indices.copy(handlers=False)
    indexes.open()
    table_names = indexes_get_table_names(indexes)
    sqls = []
    for i in indexes:
        if not (i.f_foreign_index.value and db_module.DATABASE == 'SQLITE'):
            table_name = table_names.get(i.owner_rec_id.value)
            if table_name:
                sqls.append(i.delete_index_sql(db_type, table_name))
    return sqls

def restore_indexes_sql(task):
    db_module = task.task_db_module
    db_type = task.task_db_type
    indexes = task.sys_indices.copy(handlers=False)
    indexes.open()
    table_names = indexes_get_table_names(indexes)
    sqls = []
    for i in indexes:
        if not (i.f_foreign_index.value and db_module.DATABASE == 'SQLITE'):
            table_name = table_names.get(i.owner_rec_id.value)
            if table_name:
                sqls.append(i.create_index_sql(db_type, table_name))
    return sqls



# sqlalchemy/sql/events.py
# Copyright (C) 2005-2020 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

from .base import SchemaEventTarget
from .. import event


class DDLEvents(event.Events):
    """
    Define event listeners for schema objects,
    that is, :class:`.SchemaItem` and other :class:`.SchemaEventTarget`
    subclasses, including :class:`.MetaData`, :class:`.Table`,
    :class:`.Column`.

    :class:`.MetaData` and :class:`.Table` support events
    specifically regarding when CREATE and DROP
    DDL is emitted to the database.

    Attachment events are also provided to customize
    behavior whenever a child schema element is associated
    with a parent, such as, when a :class:`.Column` is associated
    with its :class:`.Table`, when a :class:`.ForeignKeyConstraint`
    is associated with a :class:`.Table`, etc.

    Example using the ``after_create`` event::

        from sqlalchemy import event
        from sqlalchemy import Table, Column, Metadata, Integer

        m = MetaData()
        some_table = Table('some_table', m, Column('data', Integer))

        def after_create(target, connection, **kw):
            connection.execute("ALTER TABLE %s SET name=foo_%s" %
                                    (target.name, target.name))

        event.listen(some_table, "after_create", after_create)

    DDL events integrate closely with the
    :class:`.DDL` class and the :class:`.DDLElement` hierarchy
    of DDL clause constructs, which are themselves appropriate
    as listener callables::

        from sqlalchemy import DDL
        event.listen(
            some_table,
            "after_create",
            DDL("ALTER TABLE %(table)s SET name=foo_%(table)s")
        )

    The methods here define the name of an event as well
    as the names of members that are passed to listener
    functions.

    For all :class:`.DDLEvent` events, the ``propagate=True`` keyword argument
    will ensure that a given event handler is propagated to copies of the
    object, which are made when using the :meth:`.Table.tometadata` method::

        from sqlalchemy import DDL
        event.listen(
            some_table,
            "after_create",
            DDL("ALTER TABLE %(table)s SET name=foo_%(table)s"),
            propagate=True
        )

        new_table = some_table.tometadata(new_metadata)

    The above :class:`.DDL` object will also be associated with the
    :class:`.Table` object represented by ``new_table``.

    .. seealso::

        :ref:`event_toplevel`

        :class:`.DDLElement`

        :class:`.DDL`

        :ref:`schema_ddl_sequences`

    """

    _target_class_doc = "SomeSchemaClassOrObject"
    _dispatch_target = SchemaEventTarget

    def before_create(self, target, connection, **kw):
        r"""Called before CREATE statements are emitted.

        :param target: the :class:`.MetaData` or :class:`.Table`
         object which is the target of the event.
        :param connection: the :class:`.Connection` where the
         CREATE statement or statements will be emitted.
        :param \**kw: additional keyword arguments relevant
         to the event.  The contents of this dictionary
         may vary across releases, and include the
         list of tables being generated for a metadata-level
         event, the checkfirst flag, and other
         elements used by internal events.

        :func:`.event.listen` also accepts the ``propagate=True``
        modifier for this event; when True, the listener function will
        be established for any copies made of the target object,
        i.e. those copies that are generated when
        :meth:`.Table.tometadata` is used.

        """

    def after_create(self, target, connection, **kw):
        r"""Called after CREATE statements are emitted.

        :param target: the :class:`.MetaData` or :class:`.Table`
         object which is the target of the event.
        :param connection: the :class:`.Connection` where the
         CREATE statement or statements have been emitted.
        :param \**kw: additional keyword arguments relevant
         to the event.  The contents of this dictionary
         may vary across releases, and include the
         list of tables being generated for a metadata-level
         event, the checkfirst flag, and other
         elements used by internal events.

        :func:`.event.listen` also accepts the ``propagate=True``
        modifier for this event; when True, the listener function will
        be established for any copies made of the target object,
        i.e. those copies that are generated when
        :meth:`.Table.tometadata` is used.

        """

    def before_drop(self, target, connection, **kw):
        r"""Called before DROP statements are emitted.

        :param target: the :class:`.MetaData` or :class:`.Table`
         object which is the target of the event.
        :param connection: the :class:`.Connection` where the
         DROP statement or statements will be emitted.
        :param \**kw: additional keyword arguments relevant
         to the event.  The contents of this dictionary
         may vary across releases, and include the
         list of tables being generated for a metadata-level
         event, the checkfirst flag, and other
         elements used by internal events.

        :func:`.event.listen` also accepts the ``propagate=True``
        modifier for this event; when True, the listener function will
        be established for any copies made of the target object,
        i.e. those copies that are generated when
        :meth:`.Table.tometadata` is used.

        """

    def after_drop(self, target, connection, **kw):
        r"""Called after DROP statements are emitted.

        :param target: the :class:`.MetaData` or :class:`.Table`
         object which is the target of the event.
        :param connection: the :class:`.Connection` where the
         DROP statement or statements have been emitted.
        :param \**kw: additional keyword arguments relevant
         to the event.  The contents of this dictionary
         may vary across releases, and include the
         list of tables being generated for a metadata-level
         event, the checkfirst flag, and other
         elements used by internal events.

        :func:`.event.listen` also accepts the ``propagate=True``
        modifier for this event; when True, the listener function will
        be established for any copies made of the target object,
        i.e. those copies that are generated when
        :meth:`.Table.tometadata` is used.

        """

    def before_parent_attach(self, target, parent):
        """Called before a :class:`.SchemaItem` is associated with
        a parent :class:`.SchemaItem`.

        :param target: the target object
        :param parent: the parent to which the target is being attached.

        :func:`.event.listen` also accepts the ``propagate=True``
        modifier for this event; when True, the listener function will
        be established for any copies made of the target object,
        i.e. those copies that are generated when
        :meth:`.Table.tometadata` is used.

        """

    def after_parent_attach(self, target, parent):
        """Called after a :class:`.SchemaItem` is associated with
        a parent :class:`.SchemaItem`.

        :param target: the target object
        :param parent: the parent to which the target is being attached.

        :func:`.event.listen` also accepts the ``propagate=True``
        modifier for this event; when True, the listener function will
        be established for any copies made of the target object,
        i.e. those copies that are generated when
        :meth:`.Table.tometadata` is used.

        """

    def column_reflect(self, inspector, table, column_info):
        """Called for each unit of 'column info' retrieved when
        a :class:`.Table` is being reflected.

        The dictionary of column information as returned by the
        dialect is passed, and can be modified.  The dictionary
        is that returned in each element of the list returned
        by :meth:`.reflection.Inspector.get_columns`:

            * ``name`` - the column's name

            * ``type`` - the type of this column, which should be an instance
              of :class:`~sqlalchemy.types.TypeEngine`

            * ``nullable`` - boolean flag if the column is NULL or NOT NULL

            * ``default`` - the column's server default value.  This is
              normally specified as a plain string SQL expression, however the
              event can pass a :class:`.FetchedValue`, :class:`.DefaultClause`,
              or :func:`.sql.expression.text` object as well.

              .. versionchanged:: 1.1.6

                    The :meth:`.DDLEvents.column_reflect` event allows a non
                    string :class:`.FetchedValue`,
                    :func:`.sql.expression.text`, or derived object to be
                    specified as the value of ``default`` in the column
                    dictionary.

            * ``attrs``  - dict containing optional column attributes

        The event is called before any action is taken against
        this dictionary, and the contents can be modified.
        The :class:`.Column` specific arguments ``info``, ``key``,
        and ``quote`` can also be added to the dictionary and
        will be passed to the constructor of :class:`.Column`.

        Note that this event is only meaningful if either
        associated with the :class:`.Table` class across the
        board, e.g.::

            from sqlalchemy.schema import Table
            from sqlalchemy import event

            def listen_for_reflect(inspector, table, column_info):
                "receive a column_reflect event"
                # ...

            event.listen(
                    Table,
                    'column_reflect',
                    listen_for_reflect)

        ...or with a specific :class:`.Table` instance using
        the ``listeners`` argument::

            def listen_for_reflect(inspector, table, column_info):
                "receive a column_reflect event"
                # ...

            t = Table(
                'sometable',
                autoload=True,
                listeners=[
                    ('column_reflect', listen_for_reflect)
                ])

        This because the reflection process initiated by ``autoload=True``
        completes within the scope of the constructor for :class:`.Table`.

        :func:`.event.listen` also accepts the ``propagate=True``
        modifier for this event; when True, the listener function will
        be established for any copies made of the target object,
        i.e. those copies that are generated when
        :meth:`.Table.tometadata` is used.

        """



import sys
import fdb
from werkzeug._compat import text_type, to_bytes, to_unicode

DATABASE = 'FIREBIRD'
NEED_DATABASE_NAME = True
NEED_LOGIN = True
NEED_PASSWORD = True
NEED_ENCODING = True
NEED_HOST = True
NEED_PORT = True
CAN_CHANGE_TYPE = False
CAN_CHANGE_SIZE = False
DDL_ROLLBACK = True
NEED_GENERATOR = True

FROM = '"%s" AS %s'
LEFT_OUTER_JOIN = 'LEFT OUTER JOIN "%s" AS %s'
FIELD_AS = 'AS'
LIKE = 'LIKE'

JAM_TYPES = TEXT, INTEGER, FLOAT, CURRENCY, DATE, DATETIME, BOOLEAN, LONGTEXT, KEYS, FILE, IMAGE = range(1, 12)
FIELD_TYPES = {
    INTEGER: 'INTEGER',
    TEXT: 'VARCHAR',
    FLOAT: 'DOUBLE PRECISION',
    CURRENCY: 'DOUBLE PRECISION',
    DATE: 'DATE',
    DATETIME: 'TIMESTAMP',
    BOOLEAN: 'INTEGER',
    LONGTEXT: 'BLOB',
    KEYS: 'BLOB',
    FILE: 'VARCHAR(522)',
    IMAGE: 'VARCHAR(522)'
}

def connect(database, user, password, host, port, encoding, server):
    if not encoding:
        encoding = None
    if not port:
        port = None
    else:
        port = int(port)
    return fdb.connect(database=database, user=user, password=password, charset=encoding, host=host, port=port)

get_lastrowid = None

def get_select(query, fields_clause, from_clause, where_clause, group_clause, order_clause, fields):
    start = fields_clause
    end = ''.join([from_clause, where_clause, group_clause, order_clause])
    offset = query['__offset']
    limit = query['__limit']
    page = ''
    if limit:
        page = 'FIRST %d SKIP %d' % (limit, offset)
    return 'SELECT %s %s FROM %s' % (page, start, end)

def process_sql_params(params, cursor):
    result = []
    for p in params:
        if type(p) == tuple:
            value, data_type = p
            if data_type in [LONGTEXT, KEYS]:
                if type(value) == text_type:
                    value = to_bytes(value, 'utf-8')
        else:
            value = p
        result.append(value)
    return result

def process_sql_result(rows):
    result = []
    for row in rows:
        new_row = []
        for r in row:
            if isinstance(r, fdb.fbcore.BlobReader):
                r = to_unicode(r.read(), 'utf-8')
            elif type(r) == bytes:
                r = to_unicode(r, 'utf-8')
            new_row.append(r)
        result.append(new_row)
    return result

def cast_date(date_str):
    return "CAST('" + date_str + "' AS DATE)"

def cast_datetime(datetime_str):
    return "CAST('" + datetime_str + "' AS TIMESTAMP)"

def value_literal(index):
    return '?'

def convert_like(field_name, val, data_type):
    return 'UPPER(%s)' % field_name, val.upper()

def create_table_sql(table_name, fields, gen_name=None, foreign_fields=None):
    result = []
    primary_key = ''
    sql = 'CREATE TABLE "%s"\n(\n' % table_name
    lines = []
    for field in fields:
        line = '"%s" %s' % (field['field_name'], FIELD_TYPES[field['data_type']])
        if field['size'] != 0 and field['data_type'] == TEXT:
            line += '(%d)' % field['size']
        if field['default_value'] and not field['primary_key']:
            if field['data_type'] == TEXT:
                line += " DEFAULT '%s'" % field['default_value']
            else:
                line += ' DEFAULT %s' % field['default_value']
        lines.append(line)
        if field['primary_key']:
            primary_key = field['field_name']
    if primary_key:
        lines.append('CONSTRAINT %s_PR_INDEX PRIMARY KEY ("%s")\n' % \
            (table_name, primary_key))
    sql += ',\n'.join(lines)
    sql += ')\n'
    result.append(sql)
    if primary_key:
        result.append('CREATE SEQUENCE "%s"' % gen_name)
    return result

def delete_table_sql(table_name, gen_name):
    result = []
    result.append('DROP TABLE "%s"' % table_name)
    if gen_name:
        result.append('DROP SEQUENCE "%s"' % gen_name)
    return result

def create_index_sql(index_name, table_name, unique, fields, desc):
    return 'CREATE %s %s INDEX "%s" ON "%s" (%s)' % \
        (unique, desc, index_name, table_name, fields)

def create_foreign_index_sql(table_name, index_name, key, ref, primary_key):
    return 'ALTER TABLE "%s" ADD CONSTRAINT "%s" FOREIGN KEY ("%s") REFERENCES "%s"("%s")' % \
        (table_name, index_name, key, ref, primary_key)

def delete_index(table_name, index_name):
    return 'DROP INDEX "%s"' % index_name

def delete_foreign_index(table_name, index_name):
    return 'ALTER TABLE "%s" DROP CONSTRAINT "%s"' % (table_name, index_name)

def add_field_sql(table_name, field):
    result = 'ALTER TABLE "%s" ADD "%s" %s'
    result = result % (table_name, field['field_name'], FIELD_TYPES[field['data_type']])
    if field['size']:
        result += '(%d)' % field['size']
    if field['default_value']:
        if field['data_type'] == TEXT:
            result += " DEFAULT '%s'" % field['default_value']
        else:
            result += ' DEFAULT %s' % field['default_value']
    return result

def del_field_sql(table_name, field):
    return 'ALTER TABLE "%s" DROP "%s"' % (table_name, field['field_name'])

def change_field_sql(table_name, old_field, new_field):
    result = []
    if FIELD_TYPES[old_field['data_type']] != FIELD_TYPES[new_field['data_type']] \
        or old_field['size'] != new_field['size']:
        raise Exception(u"Don't know how to change field's size or type of %s, table name %s" % (old_field['field_name'], table_name))
    if old_field['field_name'] != new_field['field_name']:
        sql = 'ALTER TABLE "%s" ALTER "%s" TO "%s"' % \
            (table_name, old_field['field_name'], new_field['field_name'])
        result.append(sql)
    if old_field['default_value'] != new_field['default_value']:
        if new_field['default_value']:
            if new_field['data_type'] == TEXT:
                sql = 'ALTER TABLE "%s" ALTER "%s" SET DEFAULT' % \
                    (table_name, new_field['field_name'])
                sql +=  " '%s'" % new_field['default_value']
            else:
                sql = 'ALTER TABLE "%s" ALTER "%s" SET DEFAULT %s' % \
                    (table_name, new_field['field_name'], new_field['default_value'])
        else:
            sql = 'ALTER TABLE "%s" ALTER "%s" DROP DEFAULT' % \
                (table_name, new_field['field_name'])
        result.append(sql)
    return result

def next_sequence_value_sql(gen_name):
    return 'SELECT NEXT VALUE FOR "%s" FROM RDB$DATABASE' % gen_name

def restart_sequence_sql(gen_name, value):
    return 'ALTER SEQUENCE %s RESTART WITH %d' % (gen_name, value)

def identifier_case(name):
    return name.upper()

def get_table_names(connection):
    cursor = connection.cursor()
    cursor.execute('''
        SELECT RDB$RELATION_NAME FROM RDB$RELATIONS
        WHERE (RDB$SYSTEM_FLAG <> 1 OR RDB$SYSTEM_FLAG IS NULL)
        AND RDB$VIEW_BLR IS NULL
    ''')
    result = cursor.fetchall()
    return [r[0] for r in result]

def get_table_info(connection, table_name, db_name):
    cursor = connection.cursor()
    sql = '''
       SELECT RF.RDB$FIELD_NAME AS COLUMN_NAME,
            CASE F.RDB$FIELD_TYPE
                WHEN 7 THEN 'SMALLINT'
                WHEN 8 THEN 'INTEGER'
                WHEN 10 THEN 'FLOAT'
                WHEN 12 THEN 'DATE'
                WHEN 13 THEN 'TIME'
                WHEN 14 THEN 'CHAR'
                WHEN 16 THEN 'BIGINT'
                WHEN 27 THEN 'DOUBLE PRECISION'
                WHEN 35 THEN 'TIMESTAMP'
                WHEN 37 THEN 'VARCHAR'
                WHEN 261 THEN 'BLOB'
            END AS DATA_TYPE,
            F.RDB$FIELD_LENGTH,
            F.RDB$DEFAULT_VALUE
        FROM RDB$FIELDS F
            JOIN RDB$RELATION_FIELDS RF ON RF.RDB$FIELD_SOURCE = F.RDB$FIELD_NAME
        WHERE RF.RDB$RELATION_NAME = '%s'
    '''
    cursor.execute(sql % table_name)
    result = cursor.fetchall()
    fields = []
    for (field_name, data_type, size, default_value) in result:
        data_type = data_type.strip()
        if not data_type in ['VARCHAR', 'CHAR']:
            size = 0
        fields.append({
            'field_name': field_name.strip(),
            'data_type': data_type,
            'size': size,
            'default_value': default_value,
            'pk': False
        })
    return {'fields': fields, 'field_types': FIELD_TYPES}



import MySQLdb
from werkzeug._compat import iteritems, to_unicode

DATABASE = 'MYSQL'
NEED_DATABASE_NAME = True
NEED_LOGIN = True
NEED_PASSWORD = True
NEED_ENCODING = True
NEED_HOST = True
NEED_PORT = True
CAN_CHANGE_TYPE = False
CAN_CHANGE_SIZE = False
DDL_ROLLBACK = False
NEED_GENERATOR = False

FROM = '"%s" AS %s'
LEFT_OUTER_JOIN = 'LEFT OUTER JOIN "%s" AS %s'
FIELD_AS = 'AS'
LIKE = 'LIKE'

JAM_TYPES = TEXT, INTEGER, FLOAT, CURRENCY, DATE, DATETIME, BOOLEAN, LONGTEXT, KEYS, FILE, IMAGE = range(1, 12)
FIELD_TYPES = {
    INTEGER: 'INT',
    TEXT: 'VARCHAR',
    FLOAT: 'DOUBLE',
    CURRENCY: 'DOUBLE',
    DATE: 'DATE',
    DATETIME: 'DATETIME',
    BOOLEAN: 'INT',
    LONGTEXT: 'LONGTEXT',
    KEYS: 'LONGTEXT',
    FILE: 'LONGTEXT',
    IMAGE: 'LONGTEXT'
}

def connect(database, user, password, host, port, encoding, server):
    charset = None
    use_unicode = None
    if encoding:
        charset = encoding
        use_unicode = True
    if port:
        connection = MySQLdb.connect(db=database, user=user, passwd=password, host=host,
            port=int(port), charset=charset, use_unicode=use_unicode)
    else:
        connection = MySQLdb.connect(db=database, user=user, passwd=password, host=host,
            charset=charset, use_unicode=use_unicode)
    connection.autocommit(False)
    cursor = connection.cursor()
    cursor.execute("SET SESSION SQL_MODE=ANSI_QUOTES")
    return connection

def get_lastrowid(cursor):
    return cursor.lastrowid

def get_select(query, fields_clause, from_clause, where_clause, group_clause, order_clause, fields):
    start = fields_clause
    end = ''.join([from_clause, where_clause, group_clause, order_clause])
    offset = query['__offset']
    limit = query['__limit']
    result = 'SELECT %s FROM %s' % (start, end)
    if limit:
        result += ' LIMIT %d, %d' % (offset, limit)
    return result

def process_sql_params(params, cursor):
    result = []
    for p in params:
        if type(p) == tuple:
            value, data_type = p
        else:
            value = p
        result.append(value)
    return result

def process_sql_result(rows):
    result = []
    for row in rows:
        new_row = []
        for r in row:
            if isinstance(r, bytes):
                r = to_unicode(r, 'utf-8')
            new_row.append(r)
        result.append(new_row)
    return result


def cast_date(date_str):
    return "'" + date_str + "'"

def cast_datetime(datetime_str):
    return "'" + datetime_str + "'"

def value_literal(index):
    return '%s'

def convert_like(field_name, val, data_type):
    return field_name, val

def create_table_sql(table_name, fields, gen_name=None, foreign_fields=None):
    result = []
    primary_key = ''
    sql = 'CREATE TABLE "%s"\n(\n' % table_name
    lines = []
    for field in fields:
        line = '"%s" %s' % (field['field_name'], FIELD_TYPES[field['data_type']])
        if field['size'] != 0 and field['data_type'] == TEXT:
            line += '(%d)' % field['size']
        if field['default_value'] and not field['primary_key']:
            if field['data_type'] == TEXT:
                line += " DEFAULT '%s'" % field['default_value']
            else:
                line += ' DEFAULT %s' % field['default_value']
        if field['primary_key']:
            line += ' NOT NULL AUTO_INCREMENT'
            primary_key = field['field_name']
        lines.append(line)
    if primary_key:
        lines.append('PRIMARY KEY("%s")' % primary_key)
    sql += ',\n'.join(lines)
    sql += ')\n'
    result.append(sql)
    return result


def delete_table_sql(table_name, gen_name):
    result = []
    result.append('DROP TABLE "%s"' % table_name)
    return result

def create_index_sql(index_name, table_name, unique, fields, desc):
    return 'CREATE %s INDEX "%s" ON "%s" (%s)' % (unique, index_name, table_name, fields)

def create_foreign_index_sql(table_name, index_name, key, ref, primary_key):
    return 'ALTER TABLE "%s" ADD CONSTRAINT "%s" FOREIGN KEY ("%s") REFERENCES "%s"("%s")' % \
        (table_name, index_name, key, ref, primary_key)

def delete_index(table_name, index_name):
    return 'DROP INDEX "%s" ON "%s"' % (index_name, table_name)

def delete_foreign_index(table_name, index_name):
    return 'ALTER TABLE "%s" DROP FOREIGN KEY "%s"' % (table_name, index_name)

def add_field_sql(table_name, field):
    result = 'ALTER TABLE "%s" ADD "%s" %s' % \
        (table_name, field['field_name'], FIELD_TYPES[field['data_type']])
    if field['size']:
        result += '(%d)' % field['size']
    if field['default_value']:
        if field['data_type'] == TEXT:
            result += " DEFAULT '%s'" % field['default_value']
        else:
            result += ' DEFAULT %s' % field['default_value']
    return result

def del_field_sql(table_name, field):
    return 'ALTER TABLE "%s" DROP "%s"' % (table_name, field['field_name'])

def change_field_sql(table_name, old_field, new_field):
    result = []
    if old_field['field_name'] != new_field['field_name']:
        sql = 'ALTER TABLE "%s" CHANGE  "%s" "%s" %s' % (table_name, old_field['field_name'],
            new_field['field_name'], FIELD_TYPES[new_field['data_type']])
        if old_field['size']:
            sql += '(%d)' % old_field['size']
        #~ if old_field['data_type'] != new_field['data_type'] or \
            #~ old_field['size'] != new_field['size']:
            #~ sql += ' %s' % FIELD_TYPES[field['data_type']]
            #~ if new_field['size'] and old_field['size'] != new_field['size']:
                #~ sql += '(%d)' % new_field['size']
        result.append(sql)
    if old_field['default_value'] != new_field['default_value']:
        if new_field['default_value']:
            if new_field['data_type'] == TEXT:
                sql = 'ALTER TABLE "%s" ALTER "%s" SET DEFAULT' % \
                    (table_name, new_field['field_name'])
                sql +=  " '%s'" % new_field['default_value']
            else:
                sql = 'ALTER TABLE "%s" ALTER "%s" SET DEFAULT %s' % \
                    (table_name, new_field['field_name'], new_field['default_value'])
        else:
            sql = 'ALTER TABLE "%s" ALTER "%s" DROP DEFAULT' % \
                (table_name, new_field['field_name'])
        result.append(sql)
    return result

def param_literal():
    return '%s'

def quotes():
    return '`'

def next_sequence_value_sql(table_name):
    return None

def restart_sequence_sql(table_name, value):
    pass

def identifier_case(name):
    return name.lower()

def get_table_names(connection):
    cursor = connection.cursor()
    cursor.execute('show tables')
    result = cursor.fetchall()
    return [r[0] for r in result]

def get_table_info(connection, table_name, db_name):
    cursor = connection.cursor()
    sql = 'SHOW COLUMNS FROM "%s" FROM %s' % (table_name, db_name)
    cursor.execute(sql)
    result = cursor.fetchall()
    fields = []
    for (field_name, type_size, null, key, default_value, autoinc) in result:
        try:
            pk = False
            if autoinc and key == 'PRI':
                pk = True
            data_type = type_size.split('(')[0].upper()
            size = type_size.split('(')[1].split(')')[0]
            if not data_type in ['VARCHAR', 'CHAR']:
                size = 0
        except:
            data_type = type_size
            size = 0
        fields.append({
            'field_name': field_name,
            'data_type': data_type,
            'size': size,
            'default_value': default_value,
            'pk': pk
        })
    return {'fields': fields, 'field_types': FIELD_TYPES}




# -*- coding: utf-8 -*-
"""
    werkzeug.debug.console
    ~~~~~~~~~~~~~~~~~~~~~~

    Interactive console support.

    :copyright: 2007 Pallets
    :license: BSD-3-Clause
"""
import code
import sys
from types import CodeType

from ..local import Local
from ..utils import escape
from .repr import debug_repr
from .repr import dump
from .repr import helper


_local = Local()


class HTMLStringO(object):
    """A StringO version that HTML escapes on write."""

    def __init__(self):
        self._buffer = []

    def isatty(self):
        return False

    def close(self):
        pass

    def flush(self):
        pass

    def seek(self, n, mode=0):
        pass

    def readline(self):
        if len(self._buffer) == 0:
            return ""
        ret = self._buffer[0]
        del self._buffer[0]
        return ret

    def reset(self):
        val = "".join(self._buffer)
        del self._buffer[:]
        return val

    def _write(self, x):
        if isinstance(x, bytes):
            x = x.decode("utf-8", "replace")
        self._buffer.append(x)

    def write(self, x):
        self._write(escape(x))

    def writelines(self, x):
        self._write(escape("".join(x)))


class ThreadedStream(object):
    """Thread-local wrapper for sys.stdout for the interactive console."""

    @staticmethod
    def push():
        if not isinstance(sys.stdout, ThreadedStream):
            sys.stdout = ThreadedStream()
        _local.stream = HTMLStringO()

    @staticmethod
    def fetch():
        try:
            stream = _local.stream
        except AttributeError:
            return ""
        return stream.reset()

    @staticmethod
    def displayhook(obj):
        try:
            stream = _local.stream
        except AttributeError:
            return _displayhook(obj)
        # stream._write bypasses escaping as debug_repr is
        # already generating HTML for us.
        if obj is not None:
            _local._current_ipy.locals["_"] = obj
            stream._write(debug_repr(obj))

    def __setattr__(self, name, value):
        raise AttributeError("read only attribute %s" % name)

    def __dir__(self):
        return dir(sys.__stdout__)

    def __getattribute__(self, name):
        if name == "__members__":
            return dir(sys.__stdout__)
        try:
            stream = _local.stream
        except AttributeError:
            stream = sys.__stdout__
        return getattr(stream, name)

    def __repr__(self):
        return repr(sys.__stdout__)


# add the threaded stream as display hook
_displayhook = sys.displayhook
sys.displayhook = ThreadedStream.displayhook


class _ConsoleLoader(object):
    def __init__(self):
        self._storage = {}

    def register(self, code, source):
        self._storage[id(code)] = source
        # register code objects of wrapped functions too.
        for var in code.co_consts:
            if isinstance(var, CodeType):
                self._storage[id(var)] = source

    def get_source_by_code(self, code):
        try:
            return self._storage[id(code)]
        except KeyError:
            pass


def _wrap_compiler(console):
    compile = console.compile

    def func(source, filename, symbol):
        code = compile(source, filename, symbol)
        console.loader.register(code, source)
        return code

    console.compile = func


class _InteractiveConsole(code.InteractiveInterpreter):
    def __init__(self, globals, locals):
        locals = dict(globals)
        locals.update(locals)
        locals["dump"] = dump
        locals["help"] = helper
        locals["__loader__"] = self.loader = _ConsoleLoader()
        code.InteractiveInterpreter.__init__(self, locals)
        self.more = False
        self.buffer = []
        _wrap_compiler(self)

    def runsource(self, source):
        source = source.rstrip() + "\n"
        ThreadedStream.push()
        prompt = "... " if self.more else ">>> "
        try:
            source_to_eval = "".join(self.buffer + [source])
            if code.InteractiveInterpreter.runsource(
                self, source_to_eval, "<debugger>", "single"
            ):
                self.more = True
                self.buffer.append(source)
            else:
                self.more = False
                del self.buffer[:]
        finally:
            output = ThreadedStream.fetch()
        return prompt + escape(source) + output

    def runcode(self, code):
        try:
            exec(code, self.locals)
        except Exception:
            self.showtraceback()

    def showtraceback(self):
        from .tbtools import get_current_traceback

        tb = get_current_traceback(skip=1)
        sys.stdout._write(tb.render_summary())

    def showsyntaxerror(self, filename=None):
        from .tbtools import get_current_traceback

        tb = get_current_traceback(skip=4)
        sys.stdout._write(tb.render_summary())

    def write(self, data):
        sys.stdout.write(data)


class Console(object):
    """An interactive console."""

    def __init__(self, globals=None, locals=None):
        if locals is None:
            locals = {}
        if globals is None:
            globals = {}
        self._ipy = _InteractiveConsole(globals, locals)

    def eval(self, code):
        _local._current_ipy = self._ipy
        old_sys_stdout = sys.stdout
        try:
            return self._ipy.runsource(code)
        finally:
            sys.stdout = old_sys_stdout



import  os
import json
from werkzeug._compat import iterkeys

from ..common import consts, ProjectNotCompleted
from ..server_classes import Task, Group

def create_task(app):
    result = None
    adm = app.admin
    it = adm.sys_items.copy()
    it.set_where(type_id=consts.TASK_TYPE)
    it.open()
    if adm.task_db_type:
        result = Task(app, it.f_item_name.value, it.f_name.value,
            it.f_js_filename.value, adm.task_db_type, adm.task_db_server,
            adm.task_db_database, adm.task_db_user, adm.task_db_password,
            adm.task_db_host, adm.task_db_port, adm.task_db_encoding,
            adm.task_con_pool_size, adm.task_persist_con
            )
        result.ID = it.id.value
        load_task(result, app)
    else:
        raise ProjectNotCompleted()
    return result

def reload_task(app):
    if app.task:
        if app.task.pool is None:
            app.task.create_pool()
        load_task(app.task, app, first_build=False)

def create_fields(item, parent_id, item_dict):
    fields = item_dict['sys_fields']['item']
    fields_dict = item_dict['sys_fields']['rec_dict']
    recs = fields_dict.get(parent_id)
    if recs:
        for r in recs:
            fields.rec_no = r
            if fields.owner_rec_id.value == parent_id:
                visible = False
                word_wrap = False
                expand = False
                editable = False
                edit_visible = False
                edit_index = -1
                field = item.add_field(fields.id.value,
                    fields.f_field_name.value,
                    fields.f_name.value,
                    fields.f_data_type.value,
                    size=fields.f_size.value,
                    required=fields.f_required.value,
                    lookup_item=fields.f_object.value,
                    lookup_field=fields.f_object_field.value,
                    read_only=fields.f_read_only.value,
                    default=fields.f_default.value,
                    default_value=fields.f_default_value.value,
                    master_field=fields.f_master_field.value,
                    alignment=fields.f_alignment.value,
                    lookup_values=fields.f_lookup_values.value,
                    enable_typeahead=fields.f_enable_typehead.value,
                    field_help=fields.f_help.value,
                    field_placeholder=fields.f_placeholder.value,
                    lookup_field1=fields.f_object_field1.value,
                    lookup_field2=fields.f_object_field2.value,
                    db_field_name=fields.f_db_field_name.value,
                    field_mask=fields.f_mask.value,
                    image_edit_width=fields.f_image_edit_width.value,
                    image_edit_height=fields.f_image_edit_height.value,
                    image_view_width=fields.f_image_view_width.value,
                    image_view_height=fields.f_image_view_height.value,
                    image_placeholder=fields.f_image_placeholder.value,
                    image_camera=fields.f_image_camera.value,
                    file_download_btn=fields.f_file_download_btn.value,
                    file_open_btn=fields.f_file_open_btn.value,
                    file_accept=fields.f_file_accept.value,
                    textarea = fields.f_textarea.value,
                    do_not_sanitize = fields.f_do_not_sanitize.value
                )

def create_filters(item, parent_id, item_dict):
    filters = item_dict['sys_filters']['item']
    filters_dict = item_dict['sys_filters']['rec_dict']
    recs = filters_dict.get(parent_id)
    if recs:
        for r in recs:
            filters.rec_no = r
            if filters.owner_rec_id.value == parent_id:
                item.add_filter(
                    filters.f_filter_name.value,
                    filters.f_name.value,
                    filters.f_field.value,
                    filters.f_type.value,
                    filters.f_multi_select_all.value,
                    filters.f_data_type.value,
                    filters.f_visible.value,
                    filters.f_help.value,
                    filters.f_placeholder.value,
                    filters.id.value,
                )

def create_params(item, parent_id, item_dict):
    params = item_dict['sys_report_params']['item']
    params_dict = item_dict['sys_report_params']['rec_dict']
    recs = params_dict.get(parent_id)
    if recs:
        for r in recs:
            params.rec_no = r
            if params.owner_rec_id.value == parent_id:
                item.add_param(
                    params.f_name.value,
                    params.f_param_name.value,
                    params.f_data_type.value,
                    params.f_object.value,
                    params.f_object_field.value,
                    params.f_required.value,
                    params.f_visible.value,
                    params.f_alignment.value,
                    params.f_multi_select.value,
                    params.f_multi_select_all.value,
                    params.f_enable_typehead.value,
                    params.f_lookup_values.value,
                    params.f_help.value,
                    params.f_placeholder.value
                )

def create_items(group, item_dict):
    items = item_dict['sys_items']['item']
    items_dict = item_dict['sys_items']['rec_dict']
    recs = items_dict.get(group.ID)
    if recs:
        for r in recs:
            items.rec_no = r
            if group.type_id == consts.REPORTS_TYPE:
                add_item = group.add_report
            else:
                add_item = group.add_item
            item = add_item(
                items.f_item_name.value,
                items.f_name.value,
                items.f_table_name.value,
                items.f_visible.value,
                items.f_view_template.value,
                items.f_js_filename.value,
                items.f_soft_delete.value)
            if item:
                item.ID = items.id.value
                item.gen_name = items.f_gen_name.value
                item._virtual_table = items.f_virtual_table.value
                item.server_code = items.f_server_module.value
                item._keep_history = items.f_keep_history.value
                item.edit_lock = items.f_edit_lock.value
                item.select_all = items.f_select_all.value
                item._primary_key = items.f_primary_key.value
                item._deleted_flag = items.f_deleted_flag.value
                item._master_id = items.f_master_id.value
                item._master_rec_id = items.f_master_rec_id.value
                item._sys_id = items.sys_id.value
                if group.type_id == consts.REPORTS_TYPE:
                    create_params(item, items.id.value, item_dict)
                    item.rep_ids = []
                else:
                    items.load_interface()
                    item._view_list = items._view_list
                    item._edit_list = items._edit_list
                    create_fields(item, group.ID, item_dict)
                    create_fields(item, items.id.value, item_dict)
                    item._order_by = items._order_list
                    item.rep_ids = items._reports_list
                    create_filters(item, group.ID, item_dict)
                    create_filters(item, items.id.value, item_dict)

def create_groups(task, item_dict):
    items = item_dict['sys_items']['item']
    groups = []
    for rec in items:
        if rec.id.value == task.ID:
            task.table_name = rec.f_table_name.value
            task.template = rec.f_view_template.value
            task.js_filename = rec.f_js_filename.value
            items.load_interface()
            task.server_code = rec.f_server_module.value
        if rec.parent.value == task.ID:
            group = Group(
                task,
                task,
                rec.f_item_name.value,
                rec.f_name.value,
                rec.f_view_template.value,
                rec.f_js_filename.value,
                rec.f_visible.value,
                rec.type_id.value
            )
            group.ID = rec.id.value
            group.type_id = rec.type_id.value
            group.server_code = rec.f_server_module.value
            groups.append(group)
    for group in groups:
         create_items(group, item_dict)

def create_details(task, item_dict):
    items = item_dict['sys_items']['item']
    for it in items:
        if it.table_id.value:
            item = task.item_by_ID(it.parent.value)
            table = task.item_by_ID(it.table_id.value)
            if item and table:
                detail = item.add_detail(table)
                detail.item_name = it.f_item_name.value
                detail.ID = it.id.value
                detail.gen_name = table.gen_name
                detail.visible = it.f_visible.value
                detail.view_template = it.f_view_template.value
                detail.js_filename = it.f_js_filename.value
                detail.server_code = it.f_server_module.value
                detail.item_type = consts.ITEM_TYPES[detail.item_type_id - 1]
                items.load_interface()
                detail._view_list = items._view_list
                detail._edit_list = items._edit_list
                detail._order_by = items._order_list

def add_reports(item):
    item.reports = []
    for rep_id in item.rep_ids:
        report = item.task.item_by_ID(rep_id[0])
        if report:
            item.reports.append(report)

def process_reports(task):
    for group in task.items:
        for item in group.items:
            add_reports(item)

def process_lookup_lists(task, admin):
    lists = admin.sys_lookup_lists.copy()
    lists.open(order_by=['f_name'])
    for l in lists:
        text = l.f_lookup_values_text.value
        task.lookup_lists[l.id.value] = json.loads(l.f_lookup_values_text.value)

def fill_dict(item, parent_field_name):
    result = {}
    parent_field = item.field_by_name(parent_field_name)
    for i in item:
        d = result.get(parent_field.value)
        if d is None:
            d = []
            result[parent_field.value] = d
        d.append(i.rec_no)
    return result

def fill_rec_dicts(admin):
    result = {}
    items = [
        ['sys_items', 'f_index', 'parent'],
        ['sys_fields', 'id', 'owner_rec_id'],
        ['sys_filters', 'f_index', 'owner_rec_id'],
        ['sys_report_params', 'f_index', 'owner_rec_id']
    ]
    for item_name, order_by, parent_field_name in items:
        item = admin.item_by_name(item_name)
        copy = item.copy(handlers=False, details=False)
        copy.open(order_by=[order_by])
        result[item_name] = {}
        result[item_name]['item'] = copy
        result[item_name]['rec_dict'] = fill_dict(copy, parent_field_name)
    return result

def load_tree(admin, task):
    item_dict = fill_rec_dicts(admin)
    create_groups(task, item_dict)
    create_details(task, item_dict)
    process_reports(task)
    process_lookup_lists(task, admin)

def remove_attr(task):
    for key in list(iterkeys(task.__dict__)):
        try:
            value = task.init_dict[key]
            if hasattr(task.__dict__[key], '__call__'):
                task.__dict__[key] = value
        except:
            del task.__dict__[key]

def history_on_apply(item, delta, params):
    raise Exception('Changing of history is not allowed.')

def load_task(task, app, first_build=True, after_import=False):
    task.pool.dispose()
    task.pool.recreate()

    admin = app.admin

    remove_attr(task)
    task.items = []
    load_tree(admin, task)

    task.bind_items()
    task.compile_all()

    params = admin.sys_params.copy()
    params.open(fields=['f_history_item', 'f_lock_item'])
    task.history_item = None
    if params.f_history_item.value:
        task.history_item = task.item_by_ID(params.f_history_item.value)
        task.history_item.on_apply = history_on_apply
    if params.f_lock_item.value:
        task.lock_item = task.item_by_ID(params.f_lock_item.value)

    task.first_build = first_build
    task.after_import = after_import
    if task.on_created:
        task.on_created(task)

    internal_path = os.path.join(admin.work_dir, 'static', 'internal')
    if os.path.exists(internal_path):
        try:
            shutil.rmtree(internal_path)
        except:
            pass

    task.pool.dispose()
    task.pool = None



import os
import zipfile
import json
import datetime

from werkzeug._compat import to_unicode

from ..common import consts

metadata_items = [
    'sys_items',
    'sys_fields',
    'sys_indices',
    'sys_filters',
    'sys_report_params',
    'sys_roles',
    'sys_params',
    'sys_privileges',
    'sys_lookup_lists',
]

def export_task(task, url):
    result = {}
    result['db_type'] = task.task_db_type
    for item_name in metadata_items:
        item = task.item_by_name(item_name)
        copy = item.copy(handlers=False)
        copy.open()
        fields = []
        for field in copy.fields:
            fields.append(field.field_name)
        result[item.item_name] = {'fields': fields, 'records': copy.dataset}
    task_file = 'task.dat'
    file_name = 'task.zip'
    zip_file_name = os.path.join(task.work_dir, file_name)
    try:
        with open(task_file, 'w') as f:
            json.dump(result, f)
        with zipfile.ZipFile(zip_file_name, 'w', zipfile.ZIP_DEFLATED) as zip_file:
            zip_file.write(task_file)
            zip_dir('', zip_file, include_ext=['.html', '.js'], recursive=False)
            zip_dir('js', zip_file)
            zip_dir('css', zip_file)
            zip_dir(os.path.join('static', 'img'), zip_file)
            zip_dir(os.path.join('static', 'js'), zip_file)
            zip_dir(os.path.join('static', 'css'), zip_file)
            zip_dir(os.path.join('static', 'fonts'), zip_file)
            zip_dir(os.path.join('static', 'builder'), zip_file)
            zip_dir('utils', zip_file, exclude_ext=['.pyc'])
            zip_dir('reports', zip_file, exclude_ext=['.xml', '.ods#'], recursive=True)

        items = task.sys_items.copy()
        items.set_where(type_id=consts.TASK_TYPE)
        items.open()
        result_path = os.path.join(task.work_dir, 'static', 'internal')
        if not os.path.exists(result_path):
            os.makedirs(result_path)
        result_file = '%s_%s_%s_%s.zip' % (items.f_item_name.value, consts.VERSION,
            task.app.jam_version, datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S"))
        os.rename(to_unicode(zip_file_name, 'utf-8'), os.path.join(to_unicode(result_path, 'utf-8'), to_unicode(result_file, 'utf-8')))
        if url:
            result = '%s/static/internal/%s' % (url, result_file)
        else:
            result = result_file
    finally:
        if os.path.exists(task_file):
            os.remove(task_file)
        if os.path.exists(file_name):
            os.remove(file_name)
    return result

def zip_dir(directory, zip_file, exclude_dirs=[], include_ext=None, exclude_ext=[], recursive=True):
    folder = os.path.join(os.getcwd())
    if directory:
        folder = os.path.join(os.getcwd(), directory)
    if os.path.exists(folder):
        if recursive:
            for dirpath, dirnames, filenames in os.walk(folder):
                head, tail = os.path.split(dirpath)
                if not tail in exclude_dirs:
                    for file_name in filenames:
                        name, ext = os.path.splitext(file_name)
                        if (not include_ext or ext in include_ext) and not ext in exclude_ext:
                            file_path = os.path.join(dirpath, file_name)
                            arcname = os.path.relpath(os.path.join(directory, file_path))
                            zip_file.write(file_path, arcname)
        else:
            for file_name in os.listdir(folder):
                name, ext = os.path.splitext(file_name)
                if (not include_ext or ext in include_ext) and not ext in exclude_ext:
                    file_path = os.path.join(folder, file_name)
                    arcname = os.path.relpath(os.path.join(directory, file_path))
                    zip_file.write(file_path, arcname)



# -*- coding: utf-8 -*-
# Copyright JS Foundation and other contributors, https://js.foundation/
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, self.list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, self.list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#
# self.SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES
# LOSS OF USE, DATA, OR PROFITS OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# self.SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import, unicode_literals

from .objects import Object
from .nodes import Node
from .syntax import Syntax


class Comment(Node):
    def __init__(self, type, value, range=None, loc=None):
        self.type = type
        self.value = value
        self.range = range
        self.loc = loc


class Entry(Object):
    def __init__(self, comment, start):
        self.comment = comment
        self.start = start


class NodeInfo(Object):
    def __init__(self, node, start):
        self.node = node
        self.start = start


class CommentHandler(object):
    def __init__(self):
        self.attach = False
        self.comments = []
        self.stack = []
        self.leading = []
        self.trailing = []

    def insertInnerComments(self, node, metadata):
        #  innnerComments for properties empty block
        #  `function a(:/** comments **\/}`
        if node.type is Syntax.BlockStatement and not node.body:
            innerComments = []
            for i, entry in enumerate(self.leading):
                if metadata.end.offset >= entry.start:
                    innerComments.append(entry.comment)
                    self.leading[i] = None
                    self.trailing[i] = None
            if innerComments:
                node.innerComments = innerComments
                self.leading = [v for v in self.leading if v is not None]
                self.trailing = [v for v in self.trailing if v is not None]

    def findTrailingComments(self, metadata):
        trailingComments = []

        if self.trailing:
            for i, entry in enumerate(self.trailing):
                if entry.start >= metadata.end.offset:
                    trailingComments.append(entry.comment)
            if trailingComments:
                self.trailing = []
            return trailingComments

        last = self.stack and self.stack[-1]
        if last and last.node.trailingComments:
            firstComment = last.node.trailingComments[0]
            if firstComment and firstComment.range[0] >= metadata.end.offset:
                trailingComments = last.node.trailingComments
                del last.node.trailingComments
        return trailingComments

    def findLeadingComments(self, metadata):
        leadingComments = []

        target = None
        while self.stack:
            entry = self.stack and self.stack[-1]
            if entry and entry.start >= metadata.start.offset:
                target = entry.node
                self.stack.pop()
            else:
                break

        if target:
            if target.leadingComments:
                for i, comment in enumerate(target.leadingComments):
                    if comment.range[1] <= metadata.start.offset:
                        leadingComments.append(comment)
                        target.leadingComments[i] = None
            if leadingComments:
                target.leadingComments = [v for v in target.leadingComments if v is not None]
                if not target.leadingComments:
                    del target.leadingComments
            return leadingComments

        for i, entry in enumerate(self.leading):
            if entry.start <= metadata.start.offset:
                leadingComments.append(entry.comment)
                self.leading[i] = None
        if leadingComments:
            self.leading = [v for v in self.leading if v is not None]

        return leadingComments

    def visitNode(self, node, metadata):
        if node.type is Syntax.Program and node.body:
            return

        self.insertInnerComments(node, metadata)
        trailingComments = self.findTrailingComments(metadata)
        leadingComments = self.findLeadingComments(metadata)
        if leadingComments:
            node.leadingComments = leadingComments
        if trailingComments:
            node.trailingComments = trailingComments

        self.stack.append(NodeInfo(
            node=node,
            start=metadata.start.offset
        ))

    def visitComment(self, node, metadata):
        type = 'Line' if node.type[0] == 'L' else 'Block'
        comment = Comment(
            type=type,
            value=node.value
        )
        if node.range:
            comment.range = node.range
        if node.loc:
            comment.loc = node.loc
        self.comments.append(comment)

        if self.attach:
            entry = Entry(
                comment=Comment(
                    type=type,
                    value=node.value,
                    range=[metadata.start.offset, metadata.end.offset]
                ),
                start=metadata.start.offset
            )
            if node.loc:
                entry.comment.loc = node.loc
            node.type = type
            self.leading.append(entry)
            self.trailing.append(entry)

    def visit(self, node, metadata):
        if node.type == 'LineComment':
            self.visitComment(node, metadata)
        elif node.type == 'BlockComment':
            self.visitComment(node, metadata)
        elif self.attach:
            self.visitNode(node, metadata)



# -*- coding: utf-8 -*-
"""
    werkzeug.useragents
    ~~~~~~~~~~~~~~~~~~~

    This module provides a helper to inspect user agent strings.  This module
    is far from complete but should work for most of the currently available
    browsers.


    :copyright: 2007 Pallets
    :license: BSD-3-Clause
"""
import re


class UserAgentParser(object):
    """A simple user agent parser.  Used by the `UserAgent`."""

    platforms = (
        (" cros ", "chromeos"),
        ("iphone|ios", "iphone"),
        ("ipad", "ipad"),
        (r"darwin|mac|os\s*x", "macos"),
        ("win", "windows"),
        (r"android", "android"),
        ("netbsd", "netbsd"),
        ("openbsd", "openbsd"),
        ("freebsd", "freebsd"),
        ("dragonfly", "dragonflybsd"),
        ("(sun|i86)os", "solaris"),
        (r"x11|lin(\b|ux)?", "linux"),
        (r"nintendo\s+wii", "wii"),
        ("irix", "irix"),
        ("hp-?ux", "hpux"),
        ("aix", "aix"),
        ("sco|unix_sv", "sco"),
        ("bsd", "bsd"),
        ("amiga", "amiga"),
        ("blackberry|playbook", "blackberry"),
        ("symbian", "symbian"),
    )
    browsers = (
        ("googlebot", "google"),
        ("msnbot", "msn"),
        ("yahoo", "yahoo"),
        ("ask jeeves", "ask"),
        (r"aol|america\s+online\s+browser", "aol"),
        (r"opera|opr", "opera"),
        ("edge", "edge"),
        ("chrome|crios", "chrome"),
        ("seamonkey", "seamonkey"),
        ("firefox|firebird|phoenix|iceweasel", "firefox"),
        ("galeon", "galeon"),
        ("safari|version", "safari"),
        ("webkit", "webkit"),
        ("camino", "camino"),
        ("konqueror", "konqueror"),
        ("k-meleon", "kmeleon"),
        ("netscape", "netscape"),
        (r"msie|microsoft\s+internet\s+explorer|trident/.+? rv:", "msie"),
        ("lynx", "lynx"),
        ("links", "links"),
        ("Baiduspider", "baidu"),
        ("bingbot", "bing"),
        ("mozilla", "mozilla"),
    )

    _browser_version_re = r"(?:%s)[/\sa-z(]*(\d+[.\da-z]+)?"
    _language_re = re.compile(
        r"(?:;\s*|\s+)(\b\w{2}\b(?:-\b\w{2}\b)?)\s*;|"
        r"(?:\(|\[|;)\s*(\b\w{2}\b(?:-\b\w{2}\b)?)\s*(?:\]|\)|;)"
    )

    def __init__(self):
        self.platforms = [(b, re.compile(a, re.I)) for a, b in self.platforms]
        self.browsers = [
            (b, re.compile(self._browser_version_re % a, re.I))
            for a, b in self.browsers
        ]

    def __call__(self, user_agent):
        for platform, regex in self.platforms:  # noqa: B007
            match = regex.search(user_agent)
            if match is not None:
                break
        else:
            platform = None
        for browser, regex in self.browsers:  # noqa: B007
            match = regex.search(user_agent)
            if match is not None:
                version = match.group(1)
                break
        else:
            browser = version = None
        match = self._language_re.search(user_agent)
        if match is not None:
            language = match.group(1) or match.group(2)
        else:
            language = None
        return platform, browser, version, language


class UserAgent(object):
    """Represents a user agent.  Pass it a WSGI environment or a user agent
    string and you can inspect some of the details from the user agent
    string via the attributes.  The following attributes exist:

    .. attribute:: string

       the raw user agent string

    .. attribute:: platform

       the browser platform.  The following platforms are currently
       recognized:

       -   `aix`
       -   `amiga`
       -   `android`
       -   `blackberry`
       -   `bsd`
       -   `chromeos`
       -   `dragonflybsd`
       -   `freebsd`
       -   `hpux`
       -   `ipad`
       -   `iphone`
       -   `irix`
       -   `linux`
       -   `macos`
       -   `netbsd`
       -   `openbsd`
       -   `sco`
       -   `solaris`
       -   `symbian`
       -   `wii`
       -   `windows`

    .. attribute:: browser

        the name of the browser.  The following browsers are currently
        recognized:

        -   `aol` *
        -   `ask` *
        -   `baidu` *
        -   `bing` *
        -   `camino`
        -   `chrome`
        -   `edge`
        -   `firefox`
        -   `galeon`
        -   `google` *
        -   `kmeleon`
        -   `konqueror`
        -   `links`
        -   `lynx`
        -   `mozilla`
        -   `msie`
        -   `msn`
        -   `netscape`
        -   `opera`
        -   `safari`
        -   `seamonkey`
        -   `webkit`
        -   `yahoo` *

        (Browsers marked with a star (``*``) are crawlers.)

    .. attribute:: version

        the version of the browser

    .. attribute:: language

        the language of the browser
    """

    _parser = UserAgentParser()

    def __init__(self, environ_or_string):
        if isinstance(environ_or_string, dict):
            environ_or_string = environ_or_string.get("HTTP_USER_AGENT", "")
        self.string = environ_or_string
        self.platform, self.browser, self.version, self.language = self._parser(
            environ_or_string
        )

    def to_header(self):
        return self.string

    def __str__(self):
        return self.string

    def __nonzero__(self):
        return bool(self.browser)

    __bool__ = __nonzero__

    def __repr__(self):
        return "<%s %r/%s>" % (self.__class__.__name__, self.browser, self.version)



from .common import consts

# Client events

task_client_events = \
    {
        'on_page_loaded': 'task',
        'on_view_form_created': 'item',
        'on_edit_form_created': 'item',
        'on_filter_form_created': 'item',
        'on_param_form_created': 'report',
        'on_view_form_shown': 'item',
        'on_edit_form_shown': 'item',
        'on_filter_form_shown': 'item',
        'on_param_form_shown': 'report',
        'on_view_form_close_query': 'item',
        'on_edit_form_close_query': 'item',
        'on_filter_form_close_query': 'item',
        'on_param_form_close_query': 'report',
        'on_view_form_closed': 'item',
        'on_edit_form_closed': 'item',
        'on_filter_form_closed': 'item',
        'on_param_form_closed': 'report',
        'on_view_form_keyup': 'item, event',
        'on_view_form_keydown': 'item, event',
        'on_edit_form_keyup': 'item, event',
        'on_edit_form_keydown': 'item, event',
        'on_before_print_report': 'report'
    }

group_client_events = \
    {
        'on_view_form_created': 'item',
        'on_edit_form_created': 'item',
        'on_filter_form_created': 'item',
        'on_view_form_shown': 'item',
        'on_edit_form_shown': 'item',
        'on_filter_form_shown': 'item',
        'on_view_form_close_query': 'item',
        'on_edit_form_close_query': 'item',
        'on_filter_form_close_query': 'item',
        'on_view_form_closed': 'item',
        'on_edit_form_closed': 'item',
        'on_filter_form_closed': 'item',
        'on_param_form_closed': 'report',
        'on_view_form_keyup': 'item, event',
        'on_view_form_keydown': 'item, event',
        'on_edit_form_keyup': 'item, event',
        'on_edit_form_keydown': 'item, event'
    }

reports_client_events = \
    {
        'on_param_form_created': 'report',
        'on_param_form_shown': 'report',
        'on_param_form_close_query': 'report',
        'on_param_form_closed': 'report',
        'on_open_report': 'report, url',
        'on_before_print_report': 'report'
    }

detail_client_events = \
    {
        'on_view_form_created': 'item',
        'on_edit_form_created': 'item',
        'on_filter_form_created': 'item',
        'on_view_form_shown': 'item',
        'on_edit_form_shown': 'item',
        'on_filter_form_shown': 'item',
        'on_view_form_close_query': 'item',
        'on_edit_form_close_query': 'item',
        'on_filter_form_close_query': 'item',
        'on_view_form_closed': 'item',
        'on_edit_form_closed': 'item',
        'on_filter_form_closed': 'item',
        'on_view_form_keyup': 'item, event',
        'on_view_form_keydown': 'item, event',
        'on_edit_form_keyup': 'item, event',
        'on_edit_form_keydown': 'item, event',
        'on_before_append': 'item',
        'on_after_append': 'item',
        'on_before_edit': 'item',
        'on_after_edit': 'item',
        'on_before_delete': 'item',
        'on_after_delete': 'item',
        'on_before_cancel': 'item',
        'on_after_cancel': 'item',
        'on_before_open': 'item, params',
        'on_after_open': 'item',
        'on_before_post': 'item',
        'on_after_post': 'item',
        'on_before_scroll': 'item',
        'on_after_scroll': 'item',
        'on_filter_record': 'item',
        'on_field_changed': 'field, lookup_item',
        'on_field_select_value': 'field, lookup_item',
        'on_before_field_changed': 'field',
        'on_filters_apply': 'item',
        'on_filters_applied': 'item',
        'on_filter_changed': 'filter',
        'on_filter_select_value': 'field, lookup_item',
        'on_field_validate': 'field',
        'on_field_get_text': 'field',
        'on_field_get_html': 'field',
        'on_field_get_summary': 'field, value'
    }

item_client_events = detail_client_events
item_client_events['on_before_apply'] = 'item, params'
item_client_events['on_after_apply'] = 'item'
item_client_events['on_detail_changed'] = 'item, detail'

report_client_events = \
    {
        'on_param_form_created': 'report',
        'on_param_form_shown': 'report',
        'on_param_form_close_query': 'report',
        'on_before_print_report': 'report',
        'on_open_report': 'report, url',
        'on_param_changed': 'param, lookup_item',
        'on_param_select_value': 'param, lookup_item'
    }

# Server events

task_server_events = \
    {
        'on_created': 'task',
        'on_login': 'task, form_data, info',
        'on_open': 'item, params',
        'on_apply': 'item, delta, params, connection',
        'on_ext_request': 'task, request, params'
    }

group_server_events = \
    {
    }

reports_server_events = \
    {
        'on_convert_report': 'report'
    }

detail_server_events = \
    {
        'on_open': 'item, params'
    }

item_server_events = detail_server_events
item_server_events['on_apply'] = 'item, delta, params, connection'

report_server_events = \
    {
        'on_before_generate': 'report',
        'on_generate': 'report',
        'on_after_generate': 'report',
        'on_parsed': 'report'
    }

def get_events(item_type_id, server):
    if server:
        if item_type_id == consts.TASK_TYPE:
            return task_server_events
        elif item_type_id in [consts.ITEMS_TYPE, consts.TABLES_TYPE]:
            return group_server_events
        elif item_type_id == consts.REPORTS_TYPE:
            return reports_server_events
        elif item_type_id in [consts.ITEM_TYPE, consts.TABLE_TYPE, consts.DETAIL_TYPE]:
            return item_server_events
        elif item_type_id == consts.REPORT_TYPE:
            return report_server_events
        elif item_type_id == consts.DETAIL_TYPE:
            return detail_server_events
    else:
        if item_type_id == consts.TASK_TYPE:
            return task_client_events
        elif item_type_id in [consts.ITEMS_TYPE, consts.TABLES_TYPE]:
            return group_client_events
        elif item_type_id == consts.REPORTS_TYPE:
            return reports_client_events
        elif item_type_id in [consts.ITEM_TYPE, consts.TABLE_TYPE, consts.DETAIL_TYPE]:
            return item_client_events
        elif item_type_id == consts.REPORT_TYPE:
            return report_client_events
        elif item_type_id == consts.DETAIL_TYPE:
            return detail_client_events





"""
Sessions
========

This module contains some helper classes that help one to add session
support to a python WSGI application. For full client-side session
storage see :mod:`secure_cookie.securecookie` which implements a secure,
client-side session storage.


Application Integration
-----------------------

::

    from secure_cookie.sessions import SessionMiddleware
    from secure_cookie.sessions import FilesystemSessionStore
    app = SessionMiddleware(app, FilesystemSessionStore())

The current session will then appear in the WSGI environment as
``werkzeug.session``. However it's recommended to not use the middleware
but the stores directly in the application. However for very simple
scripts a middleware for sessions could be sufficient.

This module does not implement methods or ways to check if a session is
expired. That should be done by a cron job and storage specific. For
example to prune unused filesystem sessions one could check the modified
time of the files. If sessions are stored in the database the ``new()``
method should add an expiration timestamp for the session.

For better flexibility it's recommended to not use the middleware but
the store and session object directly in the application dispatching::

    session_store = FilesystemSessionStore()

    def application(environ, start_response):
        request = Request(environ)
        sid = request.cookies.get('cookie_name')

        if sid is None:
            request.session = session_store.new()
        else:
            request.session = session_store.get(sid)

        response = get_the_response_object(request)

        if request.session.should_save:
            session_store.save(request.session)
            response.set_cookie('cookie_name', request.session.sid)

        return response(environ, start_response)
"""
import os
import re
import tempfile
from hashlib import sha1
from os import path
from pickle import dump
from pickle import HIGHEST_PROTOCOL
from pickle import load
from random import random
from time import time

from werkzeug._compat import PY2
from werkzeug._compat import text_type
from werkzeug.datastructures import CallbackDict
from werkzeug.filesystem import get_filesystem_encoding
from werkzeug.posixemulation import rename
from werkzeug.http import dump_cookie
from werkzeug.http import parse_cookie
from werkzeug.wsgi import ClosingIterator


_sha1_re = re.compile(r"^[a-f0-9]{40}$")


def _urandom():
    if hasattr(os, "urandom"):
        return os.urandom(30)
    return text_type(random()).encode("ascii")


def generate_key(salt=None):
    if salt is None:
        salt = repr(salt).encode("ascii")
    return sha1(b"".join([salt, str(time()).encode("ascii"), _urandom()])).hexdigest()


class ModificationTrackingDict(CallbackDict):
    __slots__ = ("modified",)

    def __init__(self, *args, **kwargs):
        def on_update(self):
            self.modified = True

        self.modified = False
        super(ModificationTrackingDict, self).__init__(on_update=on_update)
        dict.update(self, *args, **kwargs)

    def copy(self):
        """Create a flat copy of the dict."""
        missing = object()
        result = object.__new__(self.__class__)
        for name in self.__slots__:
            val = getattr(self, name, missing)
            if val is not missing:
                setattr(result, name, val)
        return result

    def __copy__(self):
        return self.copy()


class Session(ModificationTrackingDict):
    """Subclass of a dict that keeps track of direct object changes.
    Changes in mutable structures are not tracked, for those you have to
    set ``modified`` to ``True`` by hand.
    """

    __slots__ = ModificationTrackingDict.__slots__ + ("sid", "new")

    def __init__(self, data, sid, new=False):
        super(Session, self).__init__(data)
        self.sid = sid
        self.new = new

    def __repr__(self):
        return "<%s %s%s>" % (
            self.__class__.__name__,
            dict.__repr__(self),
            "*" if self.should_save else "",
        )

    @property
    def should_save(self):
        """True if the session should be saved."""
        return self.modified


class SessionStore(object):
    """Base class for all session stores.

    :param session_class: The session class to use.
    """

    def __init__(self, session_class=Session):
        self.session_class = session_class

    def is_valid_key(self, key):
        """Check if a key has the correct format."""
        return _sha1_re.match(key) is not None

    def generate_key(self, salt=None):
        """Simple function that generates a new session key."""
        return generate_key(salt)

    def new(self):
        """Generate a new session."""
        return self.session_class({}, self.generate_key(), True)

    def save(self, session):
        """Save a session."""

    def save_if_modified(self, session):
        """Save if a session class wants an update."""
        if session.should_save:
            self.save(session)

    def delete(self, session):
        """Delete a session."""

    def get(self, sid):
        """Get a session for this sid or a new session object. This
        method has to check if the session key is valid and create a new
        session if that wasn't the case.
        """
        return self.session_class({}, sid, True)


# Used for temporary files by the filesystem session store.
_fs_transaction_suffix = ".__wz_sess"


class FilesystemSessionStore(SessionStore):
    """Simple example session store that saves sessions on the
    filesystem.

    :param path: The path to the folder used for storing the sessions.
        If not provided the default temporary directory is used.
    :param filename_template: A string template used to give the session
        a filename. ``%s`` is replaced with the session id.
    :param session_class: The session class to use.
    :param renew_missing: Set to ``True`` if you want the store to give
        the user a new sid if the session was not yet saved.
    """

    def __init__(
        self,
        path=None,
        filename_template="werkzeug_%s.sess",
        session_class=Session,
        renew_missing=False,
        mode=0o644,
    ):
        super(FilesystemSessionStore, self).__init__(session_class=session_class)

        if path is None:
            path = tempfile.gettempdir()

        self.path = path

        if isinstance(filename_template, text_type) and PY2:
            filename_template = filename_template.encode(get_filesystem_encoding())

        assert not filename_template.endswith(_fs_transaction_suffix), (
            "filename templates may not end with %s" % _fs_transaction_suffix
        )
        self.filename_template = filename_template
        self.renew_missing = renew_missing
        self.mode = mode

    def get_session_filename(self, sid):
        # Out of the box this should be a strict ASCII subset, but you
        # might reconfigure the session object to have a more arbitrary
        # string.
        if isinstance(sid, text_type) and PY2:
            sid = sid.encode(get_filesystem_encoding())

        return path.join(self.path, self.filename_template % sid)

    def save(self, session):
        fn = self.get_session_filename(session.sid)
        fd, tmp = tempfile.mkstemp(suffix=_fs_transaction_suffix, dir=self.path)
        f = os.fdopen(fd, "wb")

        try:
            dump(dict(session), f, HIGHEST_PROTOCOL)
        finally:
            f.close()

        try:
            rename(tmp, fn)
            os.chmod(fn, self.mode)
        except (IOError, OSError):
            pass

    def delete(self, session):
        fn = self.get_session_filename(session.sid)

        try:
            os.unlink(fn)
        except OSError:
            pass

    def get(self, sid):
        if not self.is_valid_key(sid):
            return self.new()

        try:
            f = open(self.get_session_filename(sid), "rb")
        except IOError:
            if self.renew_missing:
                return self.new()

            data = {}
        else:
            try:
                try:
                    data = load(f)
                except Exception:
                    data = {}
            finally:
                f.close()

        return self.session_class(data, sid, False)

    def list(self):
        """List all sessions in the store."""
        before, after = self.filename_template.split("%s", 1)
        filename_re = re.compile(
            r"%s(.{5,})%s$" % (re.escape(before), re.escape(after))
        )
        result = []
        for filename in os.listdir(self.path):
            #: this is a session that is still being saved.
            if filename.endswith(_fs_transaction_suffix):
                continue
            match = filename_re.match(filename)
            if match is not None:
                result.append(match.group(1))
        return result


class SessionMiddleware(object):
    """A middleware that puts the session object of a store into the
    WSGI environ. It automatically sets cookies and restores sessions.

    However a middleware is not the preferred solution because it won't
    be as fast as sessions managed by the application itself and will
    put a key into the WSGI environment only relevant for the
    application which is against the concept of WSGI.

    The cookie parameters are the same as for the :func:`~dump_cookie`
    function just prefixed with ``cookie_``. Additionally ``max_age`` is
    called ``cookie_age`` and not ``cookie_max_age`` because of
    backwards compatibility.
    """

    def __init__(
        self,
        app,
        store,
        cookie_name="session_id",
        cookie_age=None,
        cookie_expires=None,
        cookie_path="/",
        cookie_domain=None,
        cookie_secure=None,
        cookie_httponly=False,
        cookie_samesite="Lax",
        environ_key="werkzeug.session",
    ):
        self.app = app
        self.store = store
        self.cookie_name = cookie_name
        self.cookie_age = cookie_age
        self.cookie_expires = cookie_expires
        self.cookie_path = cookie_path
        self.cookie_domain = cookie_domain
        self.cookie_secure = cookie_secure
        self.cookie_httponly = cookie_httponly
        self.cookie_samesite = cookie_samesite
        self.environ_key = environ_key

    def __call__(self, environ, start_response):
        cookie = parse_cookie(environ.get("HTTP_COOKIE", ""))
        sid = cookie.get(self.cookie_name, None)

        if sid is None:
            session = self.store.new()
        else:
            session = self.store.get(sid)

        environ[self.environ_key] = session

        def injecting_start_response(status, headers, exc_info=None):
            if session.should_save:
                self.store.save(session)
                headers.append(
                    (
                        "Set-Cookie",
                        dump_cookie(
                            self.cookie_name,
                            session.sid,
                            self.cookie_age,
                            self.cookie_expires,
                            self.cookie_path,
                            self.cookie_domain,
                            self.cookie_secure,
                            self.cookie_httponly,
                            samesite=self.cookie_samesite,
                        ),
                    )
                )

            return start_response(status, headers, exc_info)

        return ClosingIterator(
            self.app(environ, injecting_start_response),
            lambda: self.store.save_if_modified(session),
        )



# engine/__init__.py
# Copyright (C) 2005-2020 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

"""SQL connections, SQL execution and high-level DB-API interface.

The engine package defines the basic components used to interface
DB-API modules with higher-level statement construction,
connection-management, execution and result contexts.  The primary
"entry point" class into this package is the Engine and its public
constructor ``create_engine()``.

"""

from . import events  # noqa
from . import util  # noqa
from .base import Connection  # noqa
from .base import Engine  # noqa
from .base import NestedTransaction  # noqa
from .base import RootTransaction  # noqa
from .base import Transaction  # noqa
from .base import TwoPhaseTransaction  # noqa
from .create import create_engine
from .create import engine_from_config
from .interfaces import Compiled  # noqa
from .interfaces import Connectable  # noqa
from .interfaces import CreateEnginePlugin  # noqa
from .interfaces import Dialect  # noqa
from .interfaces import ExceptionContext  # noqa
from .interfaces import ExecutionContext  # noqa
from .interfaces import TypeCompiler  # noqa
from .mock import create_mock_engine
from .result import BaseRow  # noqa
from .result import BufferedColumnResultProxy  # noqa
from .result import BufferedColumnRow  # noqa
from .result import BufferedRowResultProxy  # noqa
from .result import FullyBufferedResultProxy  # noqa
from .result import LegacyRow  # noqa
from .result import result_tuple  # noqa
from .result import ResultProxy  # noqa
from .result import Row  # noqa
from .result import RowMapping  # noqa
from .util import connection_memoize  # noqa
from ..sql import ddl  # noqa


__all__ = ("create_engine", "engine_from_config", "create_mock_engine")


def __go(lcls):
    from .. import future
    from . import result

    result._future_Result = future.Result


__go(locals())



# -*- coding: utf-8 -*-
# Copyright JS Foundation and other contributors, https://js.foundation/
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import, unicode_literals, print_function, division

import sys

from .esprima import parse, tokenize, Error, toDict
from . import version


def main():
    import json
    import time
    import optparse

    usage = "usage: %prog [options] [file.js]"
    parser = optparse.OptionParser(usage=usage, version=version)
    parser.add_option("--comment", dest="comment",
                      action="store_true", default=False,
                      help="Gather all line and block comments in an array")
    parser.add_option("--attachComment", dest="attachComment",
                      action="store_true", default=False,
                      help="Attach comments to nodes")
    parser.add_option("--loc", dest="loc", default=False,
                      action="store_true",
                      help="Include line-column location info for each syntax node")
    parser.add_option("--range", dest="range", default=False,
                      action="store_true",
                      help="Include index-based range for each syntax node")
    parser.add_option("--raw", dest="raw", default=False,
                      action="store_true",
                      help="Display the raw value of literals")
    parser.add_option("--tokens", dest="tokens", default=False,
                      action="store_true",
                      help="List all tokens in an array")
    parser.add_option("--tolerant", dest="tolerant", default=False,
                      action="store_true",
                      help="Tolerate errors on a best-effort basis (experimental)")
    parser.add_option("--tokenize", dest="tokenize", default=False,
                      action="store_true",
                      help="Only tokenize, do not parse.")
    parser.add_option("--module", dest="sourceType", default='string',
                      action="store_const", const='module',
                      help="Tolerate errors on a best-effort basis (experimental)")
    parser.set_defaults(jsx=True, classProperties=True)
    opts, args = parser.parse_args()

    if len(args) == 1:
        with open(args[0], 'rb') as f:
            code = f.read().decode('utf-8')
    elif sys.stdin.isatty():
        parser.print_help()
        return 64
    else:
        code = sys.stdin.read().decode('utf-8')

    options = opts.__dict__
    do_tokenize = options.pop('tokenize')

    t = time.time()
    try:
        if do_tokenize:
            del options['sourceType']
            del options['tokens']
            del options['raw']
            del options['jsx']
            res = toDict(tokenize(code, options=options))
        else:
            res = toDict(parse(code, options=options))
    except Error as e:
        res = e.toDict()
    dt = time.time() - t + 0.000000001

    print(json.dumps(res, indent=4))
    print()
    print('Parsed everyting in', round(dt, 5), 'seconds.')
    print('Thats %d characters per second' % (len(code) // dt))

    return 0


if __name__ == '__main__':
    retval = main()
    sys.exit(retval)



import sys
import cx_Oracle
from werkzeug._compat import text_type, to_bytes, to_unicode

DATABASE = 'ORACLE'
NEED_DATABASE_NAME = True
NEED_LOGIN = True
NEED_PASSWORD = True
NEED_ENCODING = False
NEED_HOST = False
NEED_PORT = False
CAN_CHANGE_TYPE = False
CAN_CHANGE_SIZE = False
DDL_ROLLBACK = False
NEED_GENERATOR = True

FROM = '"%s" %s '
LEFT_OUTER_JOIN = 'LEFT OUTER JOIN "%s" %s'
FIELD_AS = 'AS'
LIKE = 'LIKE'
DESC = 'DESC NULLS LAST'

JAM_TYPES = TEXT, INTEGER, FLOAT, CURRENCY, DATE, DATETIME, BOOLEAN, LONGTEXT, KEYS, FILE, IMAGE = range(1, 12)
FIELD_TYPES = {
    INTEGER: 'NUMBER',
    TEXT: 'VARCHAR2',
    FLOAT: 'DOUBLE PRECISION',
    CURRENCY: 'DOUBLE PRECISION',
    DATE: 'DATE',
    DATETIME: 'TIMESTAMP',
    BOOLEAN: 'NUMBER',
    LONGTEXT: 'CLOB',
    KEYS: 'CLOB',
    FILE: 'CLOB',
    IMAGE: 'CLOB'
}

def connect(database, user, password, host, port, encoding, server):
    if database and user and password:
        return cx_Oracle.connect(user=user, password=password, dsn=database)
    elif database:
        return cx_Oracle.connect(dsn=database)

get_lastrowid = None

def get_fields(query, fields, alias):
    sql = ''
    for field in fields:
        if field.master_field:
            pass
        else:
            sql += '%s."%s", ' % (alias, field.db_field_name)
    if query['__expanded']:
        for field in fields:
            if field.lookup_item:
                sql += '%s_LOOKUP, ' % field.db_field_name
    sql = sql[:-2]
    return sql

def get_select(query, fields_clause, from_clause, where_clause, group_clause, order_clause, fields):
    start = fields_clause
    end = ''.join([from_clause, where_clause, group_clause, order_clause])
    offset = query['__offset']
    limit = query['__limit']
    result = 'SELECT %s FROM %s' % (start, end)
    if limit:
        flds = get_fields(query, fields, 'b')
        rnum = offset + 1
        rownum = offset + limit
        if offset == 0:
            rnum = 0
        result = "SELECT %s FROM (SELECT a.*, rownum rnum FROM (%s) a WHERE rownum <= %s) b WHERE rnum >= %s" % \
            (flds, result, rownum, rnum)
    return result

def process_sql_params(params, cursor):
    result = []
    for i, p in enumerate(params):
        if type(p) == tuple:
            value, data_type = p
        else:
            value = p
        result.append(value)
    return result

def process_sql_result(rows):
    result = []
    for row in rows:
        fields = []
        for field in row:
            if isinstance(field, cx_Oracle.LOB):
                field = field.read()
                field = to_unicode(field, 'utf-8')
            fields.append(field)
        result.append(fields)
    return result

def cast_date(date_str):
    return "TO_DATE('" + date_str + "', 'YYYY-MM-DD')"

def cast_datetime(datetime_str):
    return "TO_DATE('" + date_str + "', 'YYYY-MM-DD  HH24:MI')"

def value_literal(index):
    return ':f%d' % index

def convert_like(field_name, val, data_type):
    if data_type in [INTEGER, FLOAT, CURRENCY]:
        return 'TO_CHAR(%s, 99999999999990.999999999999)' % field_name, val
    else:
        return field_name, val

def create_table_sql(table_name, fields, gen_name=None, foreign_fields=None):
    result = []
    primary_key = ''
    sql = 'CREATE TABLE "%s"\n(\n' % table_name
    lines = []
    for field in fields:
        line = '"%s" %s' % (field['field_name'], FIELD_TYPES[field['data_type']])
        if field['size'] != 0 and field['data_type'] == TEXT:
            line += '(%d)' % field['size']
        if field['default_value'] and not field['primary_key']:
            if field['data_type'] == TEXT:
                line += " DEFAULT '%s'" % field['default_value']
            else:
                line += ' DEFAULT %s' % field['default_value']
        if field['primary_key']:
            primary_key = field['field_name']
        lines.append(line)
    if primary_key:
        lines.append('CONSTRAINT %s_PR_INDEX PRIMARY KEY ("%s")\n' % \
            (table_name, primary_key))
    sql += ',\n'.join(lines)
    sql += ')\n'
    result.append(sql)
    if primary_key:
        result.append('CREATE SEQUENCE "%s"' % gen_name)
    return result

def delete_table_sql(table_name, gen_name):
    result = []
    result.append('DROP TABLE "%s"' % table_name)
    if gen_name:
        result.append('DROP SEQUENCE "%s"' % gen_name)
    return result

def create_index_sql(index_name, table_name, unique, fields, desc):
    return 'CREATE %s INDEX "%s" ON "%s" (%s)' % \
        (unique, index_name, table_name, fields)

def create_foreign_index_sql(table_name, index_name, key, ref, primary_key):
    return 'ALTER TABLE "%s" ADD CONSTRAINT "%s" FOREIGN KEY ("%s") REFERENCES "%s"("%s")' % \
        (table_name, index_name, key, ref, primary_key)

def delete_index(table_name, index_name):
    return 'DROP INDEX "%s"' % index_name

def delete_foreign_index(table_name, index_name):
    return 'ALTER TABLE "%s" DROP CONSTRAINT "%s"' % (table_name, index_name)

def add_field_sql(table_name, field):
    result = 'ALTER TABLE "%s" ADD "%s" %s'
    result = result % (table_name, field['field_name'], FIELD_TYPES[field['data_type']])
    if field['size']:
        result += '(%d)' % field['size']
    if field['default_value']:
        if field['data_type'] == TEXT:
            result += " DEFAULT '%s'" % field['default_value']
        else:
            result += ' DEFAULT %s' % field['default_value']
    return result

def del_field_sql(table_name, field):
    return 'ALTER TABLE "%s" DROP COLUMN "%s"' % (table_name, field['field_name'])

def change_field_sql(table_name, old_field, new_field):
    result = []
    if FIELD_TYPES[old_field['data_type']] != FIELD_TYPES[new_field['data_type']] \
        or old_field['size'] != new_field['size']:
        raise Exception(u"Don't know how to change field's size or type of %s" % old_field['field_name'])
    if old_field['field_name'] != new_field['field_name']:
        sql = 'ALTER TABLE "%s" RENAME COLUMN "%s" TO "%s"' % \
            (table_name, old_field['field_name'], new_field['field_name'])
        result.append(sql)
    if old_field['default_value'] != new_field['default_value']:
        if new_field['default_value']:
            if new_field['data_type'] == TEXT:
                sql = 'ALTER TABLE "%s" MODIFY "%s" DEFAULT' % \
                    (table_name, new_field['field_name'])
                sql +=  " '%s'" % new_field['default_value']
            else:
                sql = 'ALTER TABLE "%s" MODIFY "%s" DEFAULT %s' % \
                    (table_name, new_field['field_name'], new_field['default_value'])
        else:
            sql = 'ALTER TABLE "%s" MODIFY "%s" DEFAULT %s' % \
                (table_name, new_field['field_name'], 'NULL')
        result.append(sql)
    return result

def param_literal():
    return '?'

def next_sequence_value_sql(gen_name):
    return 'SELECT "%s".NEXTVAL FROM DUAL' % gen_name

def restart_sequence_sql(gen_name, value):
    result = []
    result.append('DROP SEQUENCE "%s"' % gen_name)
    result.append('CREATE SEQUENCE "%s" START WITH %s' % (gen_name, value))
    return result

def identifier_case(name):
    return name.upper()

def get_table_names(connection):
    cursor = connection.cursor()
    cursor.execute('SELECT table_name FROM user_tables')
    result = cursor.fetchall()
    return [r[0] for r in result]

def get_table_info(connection, table_name, db_name):
    cursor = connection.cursor()
    sql = "SELECT COLUMN_NAME, DATA_TYPE, CHAR_LENGTH, DATA_DEFAULT FROM USER_TAB_COLUMNS WHERE TABLE_NAME='%s'" % table_name
    cursor.execute(sql)
    result = cursor.fetchall()
    fields = []
    for (field_name, data_type, size, default_value) in result:
        fields.append({
            'field_name': field_name,
            'data_type': data_type,
            'size': size,
            'default_value': default_value,
            'pk': False
        })
    return {'fields': fields, 'field_types': FIELD_TYPES}
